from datetime import datetime
from typing import List, Any
from unittest import TestCase
from math import isnan
from pandas import DataFrame
from monetdbe import connect, Timestamp


def _connect(values: List[Any], type: str) -> DataFrame:
    with connect(autocommit=True) as con:
        cur = con.execute(f"create table example(d {type})")
        cur.executemany("insert into example(d) values (?)", ((v,) for v in values))
        cur.execute("select * from example")
        return cur.fetchdf()


class TestDataFrame(TestCase):
    def test_timestamp(self):
        now = datetime.now().replace(microsecond=0)  # monetdb doesn't support microseconds
        values = [
            now,
            Timestamp(2004, 2, 14, 7, 15, 0, 510000),
        ]
        df = _connect(values, 'timestamp')
        self.assertEqual(values, list(df['d']))

    def test_int(self):
        values = [5, 10, -100]
        df = _connect(values, 'int')
        self.assertEqual(values, list(df['d']))

    def test_float(self):
        values = [5.0, 10.0, -100.0, float('nan')]
        df = _connect(values, 'float')
        self.assertEqual(values[:-1], list(df['d'])[:-1])
        self.assertTrue(isnan(df['d'].iloc[-1]))

    def test_char(self):
        values = ['a', 'i', 'é']
        df = _connect(values, 'char')
        self.assertEqual(values, list(df['d']))

    def test_string(self):
        values = ['asssssssssssssssss', 'iwwwwwwwwwwwwwww', 'éooooooooooooooooooooo']
        df = _connect(values, 'string')
        self.assertEqual(values, list(df['d']))

    def test_varchar(self):
        values = ['a', 'aa', 'éooooooooooooooooooooo']
        df = _connect(values, 'string')
        self.assertEqual(values, list(df['d']))

    def test_uuid(self):
        values = ['6c49869d-45dc-4b00-ae55-5bd363c0c72c', '2ad49a96-ba10-11ea-b3de-0242ac130004']
        df = _connect(values, 'uuid')
        self.assertEqual(values, list(df['d']))




"""

"""
import collections.abc
from typing import Union, Generator, Optional, Any, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from monetdbe import Cursor  # type: ignore[attr-defined]


class Row:
    """
    A Row is a non-standard SQLite compatible data container. It tries to mimic a tuple in most of its features.

    It supports mapping access by column name and index, iteration, representation, equality testing and len().

    If two Row objects have exactly the same columns and their members are equal, they compare equal.
    """

    def __init__(self, cur: 'Cursor', row: Union[tuple, Generator[Optional[Any], Any, None]]):
        from monetdbe import Cursor
        if type(cur) != Cursor:
            raise TypeError("You need to supply a subclass of Cursor as a cursor.")

        self._cur = cur
        self._row = tuple(row)

        if hasattr(self._cur, "description") and self._cur.description:
            self._keys = tuple(i.name for i in self._cur.description)
        else:
            self._keys = tuple()

        self._key_map = dict(zip(self._keys, range(len(self._keys))))

    def __hash__(self):
        return hash(self._keys) ^ hash(self._keys)

    def __eq__(self, other) -> bool:
        if isinstance(other, type(self)):
            a = self._keys == other._keys
            b = self._row == other._row
            return a & b
        else:
            return False

    def __iter__(self):
        # return zip(self.cur.description.__iter__(), self.row.__iter__())
        return self._row.__iter__()

    def __len__(self):
        return len(self._row)

    def __getitem__(self, item):
        if isinstance(item, (int, slice)):
            return self._row.__getitem__(item)
        if isinstance(item, str):
            try:
                return self._row.__getitem__(self._key_map[item])
            except KeyError:
                raise IndexError from None
        raise TypeError(f"type {type(item)} not supported")

    def keys(self) -> Tuple[Any, ...]:
        return self._keys


collections.abc.Sequence.register(Row)



# -*- coding: iso-8859-1 -*-
# monetdbe/test/transactions.py: tests transactions
#
# Copyright (C) 2005-2007 Gerhard H�ring <gh@ghaering.de>
#
# This file is part of pymonetdbe.
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import os
import unittest
from tempfile import TemporaryDirectory

import monetdbe as monetdbe


class TransactionTests(unittest.TestCase):
    def setUp(self):
        self.db_path = TemporaryDirectory().name

        try:
            os.remove(self.db_path)
        except OSError:
            pass

        self.con1 = monetdbe.connect(self.db_path, timeout=10)
        self.cur1 = self.con1.cursor()

        self.con2 = monetdbe.connect(self.db_path, timeout=10)
        self.cur2 = self.con2.cursor()

    def tearDown(self):
        self.cur1.close()
        self.con1.close()

        self.cur2.close()
        self.con2.close()

        try:
            os.unlink(self.db_path)
        except OSError:
            pass

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_DMLDoesNotAutoCommitBefore(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        self.cur1.execute("create table test2(j int)")
        self.cur2.execute("select i from test")
        res = self.cur2.fetchall()
        self.assertEqual(len(res), 0)

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_InsertStartsTransaction(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        self.cur2.execute("select i from test")
        res = self.cur2.fetchall()
        self.assertEqual(len(res), 0)

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_UpdateStartsTransaction(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        self.con1.commit()
        self.cur1.execute("update test set i=6")
        self.cur2.execute("select i from test")
        res = self.cur2.fetchone()[0]
        self.assertEqual(res, 5)

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_DeleteStartsTransaction(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        self.con1.commit()
        self.cur1.execute("delete from test")
        self.cur2.execute("select i from test")
        res = self.cur2.fetchall()
        self.assertEqual(len(res), 1)

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_ReplaceStartsTransaction(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        self.con1.commit()
        self.cur1.execute("replace into test(i) values (6)")
        self.cur2.execute("select i from test")
        res = self.cur2.fetchall()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0][0], 5)

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_ToggleAutoCommit(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        self.con1.isolation_level = None
        self.assertEqual(self.con1.isolation_level, None)
        self.cur2.execute("select i from test")
        res = self.cur2.fetchall()
        self.assertEqual(len(res), 1)

        self.con1.isolation_level = "DEFERRED"
        self.assertEqual(self.con1.isolation_level, "DEFERRED")
        self.cur1.execute("insert into test(i) values (5)")
        self.cur2.execute("select i from test")
        res = self.cur2.fetchall()
        self.assertEqual(len(res), 1)

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_RaiseTimeout(self):
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        with self.assertRaises(monetdbe.OperationalError):
            self.cur2.execute("insert into test(i) values (5)")

    @unittest.skip("disabled: for MonetDBe both cursors are equal")
    def test_Locking(self):
        """
        This tests the improved concurrency with pymonetdbe 2.3.4. You needed
        to roll back con2 before you could commit con1.
        """
        self.cur1.execute("create table test(i int)")
        self.cur1.execute("insert into test(i) values (5)")
        with self.assertRaises(monetdbe.OperationalError):
            self.cur2.execute("insert into test(i) values (5)")
        # NO self.con2.rollback() HERE!!!
        self.con1.commit()

    @unittest.skip("disabled: MonetDBe does not yet support concurrent connections to different dbfarm's")
    def test_RollbackCursorConsistency(self):
        """
        Checks if cursors on the connection are set into a "reset" state
        when a rollback is done on the connection.
        """
        con = monetdbe.connect(":memory:")
        cur = con.cursor()
        cur.execute("create table test(x int)")
        cur.execute("insert into test(x) values (5)")
        cur.execute("select 1 union select 2 union select 3")

        con.rollback()
        with self.assertRaises(monetdbe.InterfaceError):
            cur.fetchall()
        con.close()


class SpecialCommandTests(unittest.TestCase):
    def setUp(self):
        self.con = monetdbe.connect(":memory:")
        self.cur = self.con.cursor()

    def test_DropTable(self):
        # note (gijs): added int type
        self.cur.execute("create table test(i int)")
        self.cur.execute("insert into test(i) values (5)")
        self.cur.execute("drop table test")

    @unittest.skip("disabled, we don't support pragma")
    def test_Pragma(self):
        # note (gijs): added int type
        self.cur.execute("create table test(i int)")
        self.cur.execute("insert into test(i) values (5)")
        self.cur.execute("pragma count_changes=1")

    def tearDown(self):
        self.cur.close()
        self.con.close()


class TransactionalDDL(unittest.TestCase):
    def setUp(self):
        self.con = monetdbe.connect(":memory:")

    @unittest.skip("monetdbe does not support this")
    def test_DdlDoesNotAutostartTransaction(self):
        # For backwards compatibility reasons, DDL statements should not
        # implicitly start a transaction.
        self.con.execute("create table test(i int)")
        self.con.rollback()
        result = self.con.execute("select * from test").fetchall()
        self.assertEqual(result, [])

    @unittest.skip("This is now a duplicate of test below")
    def test_ImmediateTransactionalDDL(self):
        # You can achieve transactional DDL by issuing a BEGIN
        # statement manually.
        self.con.execute("begin transaction")
        self.con.execute("create table test(i int)")
        self.con.rollback()
        with self.assertRaises(monetdbe.OperationalError):
            self.con.execute("select * from test")

    def test_TransactionalDDL(self):
        # You can achieve transactional DDL by issuing a BEGIN
        # statement manually.
        # self.con.execute("begin transaction") - disabled, we don';t support nested transactions
        self.con.execute("create table test(i int)")
        self.con.rollback()
        with self.assertRaises(monetdbe.OperationalError):
            self.con.execute("select * from test")

    def tearDown(self):
        self.con.close()


def suite():
    default_suite = unittest.makeSuite(TransactionTests, "Check")
    special_command_suite = unittest.makeSuite(SpecialCommandTests, "Check")
    ddl_suite = unittest.makeSuite(TransactionalDDL, "Check")
    return unittest.TestSuite((default_suite, special_command_suite, ddl_suite))


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



import numpy
import pytest


class TestMultipleResultSets:
    def test_regular_selection(self, monetdbe_cursor):
        monetdbe_cursor.execute('SELECT * FROM integers')
        monetdbe_cursor.execute('SELECT * FROM integers')
        result = monetdbe_cursor.fetchall()

        expected = [(i,) for i in range(10)] + [(None,)]
        assert result == expected, "Incorrect result returned"

    def test_numpy_selection(self, monetdbe_cursor):
        monetdbe_cursor.execute('SELECT * FROM integers')
        monetdbe_cursor.execute('SELECT * FROM integers')
        result = monetdbe_cursor.fetchnumpy()
        expected = numpy.ma.masked_array(numpy.arange(11), mask=([False] * 10 + [True]))

        numpy.testing.assert_array_equal(result['i'], expected)



import logging
from _warnings import warn
from pathlib import Path
from typing import Optional, Tuple, Any, Mapping, Iterator, Dict, TYPE_CHECKING
from decimal import Decimal
from collections import namedtuple

import numpy as np

from monetdbe._lowlevel import ffi, lib
from monetdbe import exceptions
from monetdbe._cffi.convert import make_string, monet_c_type_map, extract, numpy_monetdb_map
from monetdbe._cffi.convert.bind import prepare_bind
from monetdbe._cffi.convert.bind import monetdbe_decimal_to_bte, monetdbe_decimal_to_sht, monetdbe_decimal_to_int, monetdbe_decimal_to_lng

from monetdbe._cffi.errors import check_error
from monetdbe._cffi.types_ import monetdbe_result, monetdbe_database, monetdbe_column, monetdbe_statement

if TYPE_CHECKING:
    from monetdbe.connection import Connection

_logger = logging.getLogger(__name__)


def result_fetch(result: monetdbe_result, column: int) -> monetdbe_column:
    p_rcol = ffi.new("monetdbe_column **")
    check_error(lib.monetdbe_result_fetch(result, p_rcol, column))
    return p_rcol[0]


def result_fetch_numpy(result: monetdbe_result) -> Mapping[str, np.ndarray]:
    result_dict: Dict[str, np.ndarray] = {}
    for c in range(result.ncols):
        rcol = result_fetch(result, c)
        name = make_string(rcol.name)
        type_info = monet_c_type_map[rcol.type]

        # for non float/int we for now first make a numpy object array which we then convert to the right numpy type
        if type_info.numpy_type.type == np.object_:
            np_col: np.ndarray = np.array([extract(rcol, r) for r in range(result.nrows)])
            if rcol.type == lib.monetdbe_str:
                np_col = np_col.astype(str)
            elif rcol.type == lib.monetdbe_date:
                np_col = np_col.astype('datetime64[D]')  # type: ignore
            elif rcol.type == lib.monetdbe_time:
                warn("Not converting column with type column since no proper numpy equivalent")
            elif rcol.type == lib.monetdbe_timestamp:
                np_col = np_col.astype('datetime64[ns]')  # type: ignore
        else:
            buffer_size = result.nrows * type_info.numpy_type.itemsize  # type: ignore
            c_buffer = ffi.buffer(rcol.data, buffer_size)
            np_col = np.frombuffer(c_buffer, dtype=type_info.numpy_type)  # type: ignore

        if type_info.null_value:
            mask = np_col == type_info.null_value
        else:
            mask = np.ma.nomask  # type: ignore[attr-defined]

        masked: np.ndarray = np.ma.masked_array(np_col, mask=mask)

        result_dict[name] = masked
    return result_dict


def get_autocommit() -> bool:
    value = ffi.new("int *")
    check_error(lib.monetdbe_get_autocommit(value))
    return bool(value[0])


TypeInfo = namedtuple('TypeInfo', ('impl_type', 'sql_type', 'scale'))


def bind(statement: monetdbe_statement, data: Any, parameter_nr: int, type_info=None) -> None:
    try:
        _type_info = type_info[parameter_nr]
        if (_type_info.sql_type == 'decimal'):
            d = int(Decimal(data) * (Decimal(10) ** _type_info.scale))
            if (_type_info.impl_type == 'bte'):
                prepared = monetdbe_decimal_to_bte(d)
            elif (_type_info.impl_type == 'sht'):
                prepared = monetdbe_decimal_to_sht(d)
            elif (_type_info.impl_type == 'int'):
                prepared = monetdbe_decimal_to_int(d)
            elif (_type_info.impl_type == 'lng'):
                prepared = monetdbe_decimal_to_lng(d)
            else:
                raise NotImplementedError("Unknown decimal implementation type")
        else:
            prepared = prepare_bind(data)
    except IndexError as e:
        from monetdbe import exceptions
        raise exceptions.ProgrammingError from e
    check_error(lib.monetdbe_bind(statement, prepared, parameter_nr))


def execute(statement: monetdbe_statement, make_result: bool = False) -> Tuple[monetdbe_result, int]:
    if make_result:
        p_result = ffi.new("monetdbe_result **")
    else:
        p_result = ffi.NULL

    affected_rows = ffi.new("monetdbe_cnt *")
    check_error(lib.monetdbe_execute(statement, p_result, affected_rows))

    if make_result:
        result = p_result[0]
    else:
        result = None

    return result, affected_rows[0]


def version() -> str:
    return ffi.string(lib.monetdbe_version()).decode()


class Internal:
    _active_context: Optional['Internal'] = None
    _active_connection: Optional['Connection'] = None
    _monetdbe_database: Optional[monetdbe_database] = None

    def __init__(
            self,
            connection: "Connection",
            dbdir: Optional[Path] = None,
            memorylimit: int = 0,
            querytimeout: int = 0,
            sessiontimeout: int = 0,
            nr_threads: int = 0,
            have_hge: bool = False
    ):
        self._connection = connection
        self.dbdir = dbdir
        self.memorylimit = memorylimit
        self.querytimeout = querytimeout
        self.sessiontimeout = sessiontimeout
        self.nr_threads = nr_threads
        self.have_hge = have_hge
        self._switch()
        self._monetdbe_database = self.open()

    @classmethod
    def set_active_context(cls, active_context: Optional['Internal']):
        cls._active_context = active_context

    def set_monetdbe_database(self, connection: Optional[monetdbe_database]):
        self._monetdbe_database = connection

    def __del__(self):
        if self._active_context == self:
            # only close if we are deleting the active context
            self.close()

    def _switch(self):
        if self._active_context == self:
            return

        self.set_active_context(self)

    def cleanup_result(self, result: monetdbe_result):
        self._switch()
        _logger.info("cleanup_result called")
        if result and self._monetdbe_database:
            check_error(lib.monetdbe_cleanup_result(self._monetdbe_database, result))

    def open(self) -> monetdbe_database:

        if not self.dbdir:
            url = ffi.NULL
        else:
            url = str(self.dbdir.resolve().absolute()).encode()

        p_monetdbe_database = ffi.new("monetdbe_database *")

        p_options = ffi.new("monetdbe_options *")
        p_options.memorylimit = self.memorylimit
        p_options.querytimeout = self.querytimeout
        p_options.sessiontimeout = self.sessiontimeout
        p_options.nr_threads = self.nr_threads

        result_code = lib.monetdbe_open(p_monetdbe_database, url, p_options)
        connection = p_monetdbe_database[0]

        errors = {
            0: "OK",
            -1: "Allocation failed",
            -2: "Error in DB",
        }

        if result_code:
            if result_code == -2:
                error = ffi.string(lib.monetdbe_error(connection)).decode()
                lib.monetdbe_close(connection)
            else:
                error = errors.get(result_code, "unknown error")
            msg = f"Failed to open database: {error} (code {result_code})"
            raise exceptions.OperationalError(msg)

        return connection

    def close(self) -> None:
        self._switch()
        if self._monetdbe_database:
            if lib.monetdbe_close(self._monetdbe_database):
                raise exceptions.OperationalError("Failed to close database")
            self.set_monetdbe_database(None)

        if self._active_context:
            self.set_active_context(None)

    def query(self, query: str, make_result: bool = False) -> Tuple[Optional[Any], int]:
        """
        Execute a query.

        Args:
            query: the query
            make_result: Create and return a result object. If enabled, you need to call cleanup_result on the
                          result afterwards

        returns:
            result, affected_rows

        """
        self._switch()
        if make_result:
            p_result = ffi.new("monetdbe_result **")
        else:
            p_result = ffi.NULL

        affected_rows = ffi.new("monetdbe_cnt *")
        check_error(lib.monetdbe_query(self._monetdbe_database, query.encode(), p_result, affected_rows))

        if make_result:
            result = p_result[0]
        else:
            result = None

        return result, affected_rows[0]

    def set_autocommit(self, value: bool) -> None:
        self._switch()
        check_error(lib.monetdbe_set_autocommit(self._monetdbe_database, int(value)))

    def in_transaction(self) -> bool:
        self._switch()
        return bool(lib.monetdbe_in_transaction(self._monetdbe_database))

    def append(self, table: str, data: Mapping[str, np.ndarray], schema: str = 'sys') -> None:
        """
        Directly append an array structure
        """
        self._switch()
        n_columns = len(data)
        existing_columns = list(self.get_columns(schema=schema, table=table))
        existing_names, existing_types = zip(*existing_columns)
        if not set(existing_names) == set(data.keys()):
            error = f"Appended column names ({', '.join(str(i) for i in data.keys())}) " \
                    f"don't match existing column names ({', '.join(existing_names)})"
            raise exceptions.ProgrammingError(error)

        work_columns = ffi.new(f'monetdbe_column * [{n_columns}]')
        work_objs = []
        for column_num, (column_name, existing_type) in enumerate(existing_columns):
            column_values = data[column_name]
            work_column = ffi.new('monetdbe_column *')
            type_info = numpy_monetdb_map(column_values.dtype)
            if not type_info.c_type == existing_type:
                existing_type_string = monet_c_type_map[existing_type].c_string_type
                error = f"Type '{type_info.c_string_type}' for appended column '{column_name}' " \
                        f"does not match table type '{existing_type_string}'"
                raise exceptions.ProgrammingError(error)
            work_column.type = type_info.c_type
            work_column.count = column_values.shape[0]
            work_column.name = ffi.new('char[]', column_name.encode())
            work_column.data = ffi.cast(f"{type_info.c_string_type} *", ffi.from_buffer(column_values))
            work_columns[column_num] = work_column
            work_objs.append(work_column)
        check_error(
            lib.monetdbe_append(self._monetdbe_database, schema.encode(), table.encode(), work_columns, n_columns))

    def prepare(self, query: str) -> monetdbe_statement:
        self._switch()

        stmt = ffi.new("monetdbe_statement **")
        p_result = ffi.new("monetdbe_result **")
        check_error(lib.monetdbe_prepare(self._monetdbe_database, str(query).encode(), stmt, p_result))

        input_parameter_info = list()

        for r in range(p_result[0].nrows):
            if (extract(result_fetch(p_result[0], 3), r)) is None:
                row = TypeInfo(impl_type=extract(result_fetch(p_result[0], 6), r), sql_type=extract(result_fetch(p_result[0], 0), r), scale=extract(result_fetch(p_result[0], 2), r))
                input_parameter_info.append(row)

        return stmt[0], input_parameter_info

    def cleanup_statement(self, statement: monetdbe_statement) -> None:
        self._switch()
        lib.monetdbe_cleanup_statement(self._monetdbe_database, statement)

    def dump_database(self, backupfile: Path):
        # todo (gijs): use :)
        lib.monetdbe_dump_database(self._monetdbe_database, str(backupfile).encode())

    def dump_table(self, schema_name: str, table_name: str, backupfile: Path):
        # todo (gijs): use :)
        lib.monetdbe_dump_table(self._monetdbe_database, schema_name.encode(), table_name.encode(),
                                str(backupfile).encode())

    def get_columns(self, table: str, schema: str = 'sys') -> Iterator[Tuple[str, int]]:
        self._switch()
        count_p = ffi.new('size_t*')
        columns_p = ffi.new('monetdbe_column**')

        lib.monetdbe_get_columns(self._monetdbe_database, schema.encode(), table.encode(), count_p, columns_p)

        for i in range(count_p[0]):
            name = ffi.string(columns_p[0][i].name).decode()
            type_ = columns_p[0][i].type
            yield name, type_


from monetdbe._cffi.branch import newer_then_jul2021
if not newer_then_jul2021:
    def bind(statement: monetdbe_statement, data: Any, parameter_nr: int, type_info=None) -> None:
        prepared = prepare_bind(data)
        check_error(lib.monetdbe_bind(statement, prepared, parameter_nr))

    def prepare(self, query: str) -> monetdbe_statement:
        self._switch()
        stmt = ffi.new("monetdbe_statement **")
        check_error(lib.monetdbe_prepare(self._monetdbe_database, str(query).encode(), stmt))
        return stmt[0],

    def get_columns(self, table: str, schema: str = 'sys') -> Iterator[Tuple[str, int]]:
        self._switch()
        count_p = ffi.new('size_t *')
        names_p = ffi.new('char ***')
        types_p = ffi.new('int **')

        lib.monetdbe_get_columns(self._monetdbe_database, schema.encode(), table.encode(), count_p, names_p, types_p)

        for i in range(count_p[0]):
            name = ffi.string(names_p[0][i]).decode()
            type_ = types_p[0][i]
            yield name, type_

    import sys
    Module = sys.modules[__name__]
    setattr(Module, 'bind', bind)
    setattr(Internal, 'prepare', prepare)
    setattr(Internal, 'get_columns', get_columns)



from monetdbe.monetize import monet_identifier_escape as identifier_escape
import numpy
import pytest


class TestMultipleResultSets:
    def test_string_insertion(self, monetdbe_cursor):
        monetdbe_cursor.execute('CREATE TABLE strings(s STRING)')
        monetdbe_cursor.executemany('INSERT INTO strings VALUES (%s)', [["'hello\" world\"'"]])
        monetdbe_cursor.execute('SELECT * FROM strings')
        result = monetdbe_cursor.fetchall()
        assert result == [("'hello\" world\"'",)], "Incorrect result returned"

    def test_table_name(self, monetdbe_cursor):
        sname = "table"
        tname = 'integerinios'
        monetdbe_cursor.execute('CREATE SCHEMA %s' % identifier_escape(sname))
        monetdbe_cursor.create(tname, {'i': numpy.arange(3)}, schema=sname)
        monetdbe_cursor.execute('SELECT * FROM %s.%s' % (identifier_escape(sname), identifier_escape(tname)))
        result = monetdbe_cursor.fetchall()
        assert result == [(0,), (1,), (2,)], "Incorrect result returned"



"""
For now we can't handle multiple in-memory databases simultaniously. An error should be raised when someone
tries to open multiple in memory databases.
"""
from tempfile import TemporaryDirectory
from unittest import TestCase, skip

from monetdbe import connect
from monetdbe.exceptions import OperationalError


@skip("If enabled makes the whole test suite fail. ")
class TestMultipleInmemory(TestCase):
    def test_multi_in_mem(self):
        try:
            con1 = connect(":memory:")
            with self.assertRaises(OperationalError):
                con2 = connect(":memory:")
        except Exception:
            del con1
            try:
                del con2
            except UnboundLocalError:
                ...
            raise

    def test_one_mem_one_filesystem(self):

        try:
            loc1 = TemporaryDirectory().name
            con1 = connect(loc1)
            con2 = connect(":memory:")
        except Exception:
            del con1
            del con2
            raise



"""
This is a build-time utility that will generate the monetdbe._lowlevel
CFFI bridging shared library.
"""
from pathlib import Path
from sys import platform
from os import environ
from cffi import FFI
from jinja2 import Template


monetdb_branch = environ.get("MONETDB_BRANCH", "default")
print(f"\n**MONETDB**: We are assuming you are building against MonetDB branch {monetdb_branch}")
print("**MONETDB**: If this is incorrect, set the MONETDB_BRANCH environment variable during monetdbe-python build\n")

branch_file = str(Path(__file__).parent / 'branch.py')
newer_then_jul2021 = monetdb_branch.lower() not in ("oct2020", "jul2021")

with open(branch_file, 'w') as f:
    f.write("# this file is created by the cffi interface builder and contains the monetdb branch env variable.\n\n")
    f.write(f"monetdb_branch = '{monetdb_branch}'\n")
    f.write(f"newer_then_jul2021 = {newer_then_jul2021}\n")


default = monetdb_branch.lower() in ("default", "jul2021")
win32 = platform == 'win32'


source = """
#include "monetdb/monetdbe.h"
"""


# the ffibuilder object needs to exist and be configured in the module namespace so setup.py can reach it
ffibuilder = FFI()
ffibuilder.set_source("monetdbe._lowlevel", source, libraries=['monetdbe'])
embed_path = str(Path(__file__).resolve().parent / 'embed.h.j2')


with open(embed_path, 'r') as f:
    content = f.read()
    template = Template(content)
    cdef = template.render(win32=win32, default=default, newer_then_jul2021=newer_then_jul2021)
    ffibuilder.cdef(cdef)


def build():
    ffibuilder.compile(verbose=True)


if __name__ == "__main__":
    build()



from datetime import datetime, date

PARSE_DECLTYPES = 1
PARSE_COLNAMES = 2
converters = {}
adapters = {}


def register_adapter(k, v):
    global adapters
    adapters[k] = v


def register_converter(k, v):
    global converters
    converters[k] = v


def register_adapters_and_converters():
    def adapt_date(val):
        return val.isoformat()

    def adapt_datetime(val):
        return val.isoformat(" ")

    def convert_date(val):
        return date(*map(int, val.split(b"-")))

    def convert_timestamp(val):
        datepart, timepart = val.split(b" ")
        year, month, day = map(int, datepart.split(b"-"))
        timepart_full = timepart.split(b".")
        hours, minutes, seconds = map(int, timepart_full[0].split(b":"))
        if len(timepart_full) == 2:
            microseconds = int('{:0<6.6}'.format(timepart_full[1].decode()))
        else:
            microseconds = 0

        val = datetime(year, month, day, hours, minutes, seconds, microseconds)
        return val

    register_adapter(date, adapt_date)
    register_adapter(datetime, adapt_datetime)
    register_converter("date", convert_date)
    register_converter("timestamp", convert_timestamp)


register_adapters_and_converters()

# Clean up namespace
del register_adapters_and_converters




class TestDescription(object):
    def test_description(self, monetdbe_cursor):
        monetdbe_cursor.execute('select * from sys.tables')
        assert monetdbe_cursor.description is not None

    def test_description_fields(self, monetdbe_cursor):
        monetdbe_cursor.execute('select name from sys.tables')
        assert monetdbe_cursor.description[0][0] == "name"
        assert monetdbe_cursor.description[0][1] == "string"




from typing import Dict
from logging import getLogger
from pprint import pprint

_logger = getLogger(__file__)


def get_info() -> Dict:
    """
    Fetch some MonetDBe specific properties
    """
    from monetdbe import connect
    from monetdbe._cffi.internal import version

    from_monetdb = connect().execute("select * from env()").fetchall()

    try:
        from monetdbe._cffi.branch import monetdb_branch
    except ImportError as e:
        _logger.error(f"can't import branch file: {e}")
        monetdb_branch = "not set"

    try:
        version_ = version()
    except Exception as e:
        _logger.error(f"can't call monetdb version() function: {e}")
        version_ = "not set"

    return {"from_monetdb": from_monetdb, "monetdb_branch": monetdb_branch, "version": version_}


def print_info():
    pprint(get_info())



import numpy
import pytest
from shutil import rmtree
import monetdbe


@pytest.fixture(scope="function")
def monetdbe_empty_cursor(request, tmp_path):
    test_dbfarm = tmp_path.resolve().as_posix()
    connection = monetdbe.make_connection(test_dbfarm)

    def finalizer():
        if tmp_path.is_dir():
            connection.close()
            rmtree(test_dbfarm, ignore_errors=True)

    request.addfinalizer(finalizer)
    cursor = connection.cursor()
    return cursor


@pytest.fixture(scope="function")
def monetdbe_cursor(request, tmp_path):
    test_dbfarm = tmp_path.resolve().as_posix()

    connection = monetdbe.make_connection(test_dbfarm)

    def finalizer():
        if tmp_path.is_dir():
            connection.close()
            rmtree(test_dbfarm, ignore_errors=True)

    request.addfinalizer(finalizer)

    cursor = connection.cursor()
    cursor.create('integers', {'i': numpy.arange(10)})
    cursor.execute('INSERT INTO integers VALUES(NULL)')
    return cursor


@pytest.fixture(scope="function")
def monetdbe_cursor_autocommit(tmp_path):
    test_dbfarm = tmp_path.resolve().as_posix()
    connection = monetdbe.connect(test_dbfarm)

    connection.set_autocommit(True)
    cursor = connection.cursor()

    class Context:
        def __init__(self, cursor, connection, dbfarm):
            self.cursor = cursor
            self.connection = connection
            self.dbfarm = dbfarm

    context = Context(cursor, connection, test_dbfarm)

    yield context

    if tmp_path.is_dir():
        context.connection.close()
        rmtree(context.dbfarm, ignore_errors=True)


@pytest.fixture(scope="function")
def initialize_monetdbe(request, tmp_path):
    test_dbfarm = tmp_path.resolve().as_posix()
    connection = monetdbe.connect(test_dbfarm)

    def finalizer():
        if tmp_path.is_dir():
            connection.close()
            rmtree(test_dbfarm, ignore_errors=True)

    request.addfinalizer(finalizer)
    monetdbe.connect(test_dbfarm)
    return test_dbfarm



# -*- coding: iso-8859-1 -*-
# monetdbe/test/regression.py: pymonetdbe regression tests
#
# Copyright (C) 2006-2010 Gerhard H�ring <gh@ghaering.de>
#
# This file is part of pymonetdbe.
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import datetime
import functools
import gc
import unittest
import weakref
from pathlib import Path
from shutil import rmtree

import numpy as np
import pandas as pd

import monetdbe as monetdbe
from monetdbe._cffi.branch import monetdb_branch

OCT2020 = monetdb_branch == "Oct2020"


class RegressionTests(unittest.TestCase):
    def setUp(self):
        self.con = monetdbe.connect(":memory:")

    def tearDown(self):
        self.con.close()

    # def test_PragmaUserVersion(self):
    #     # This used to crash pymonetdbe because this pragma command returns NULL for the column name
    #     cur = self.con.cursor()
    #     cur.execute("pragma user_version")
    #
    # def test_PragmaSchemaVersion(self):
    #     # This still crashed pymonetdbe <= 2.2.1
    #     con = monetdbe.connect(":memory:", detect_types=monetdbe.PARSE_COLNAMES)
    #     try:
    #         cur = self.con.cursor()
    #         cur.execute("pragma schema_version")
    #     finally:
    #         cur.close()
    #         con.close()

    @unittest.skip("cached_statements not supported (yet)")
    def test_StatementReset(self):
        # sqlite 2.1.0 to 2.2.0 have the problem that not all statements are
        # reset before a rollback, but only those that are still in the
        # statement cache. The others are not accessible from the connection object.
        con = monetdbe.connect(":memory:", cached_statements=5)
        cursors = [con.cursor() for x in range(5)]
        cursors[0].execute("create table test(x)")
        for i in range(10):
            cursors[0].executemany("insert into test(x) values (?)", [(x,) for x in range(10)])

        for i in range(5):
            cursors[i].execute(" " * i + "select x from test")

        con.rollback()

    def test_ColumnNameWithSpaces(self):
        cur = self.con.cursor()
        cur.execute('select 1 as "foo bar [datetime]"')
        self.assertEqual(cur.description[0][0], "foo bar [datetime]")

        cur.execute('select 1 as "foo baz"')
        self.assertEqual(cur.description[0][0], "foo baz")

    def test_StatementFinalizationOnCloseDb(self):
        # sqlite versions <= 2.3.3 only finalized statements in the statement
        # cache when closing the database. statements that were still
        # referenced in cursors weren't closed and could provoke "
        # "OperationalError: Unable to close due to unfinalised statements".
        cursors = []
        # default statement cache size is 100
        con = self.con
        for i in range(105):
            cur = con.cursor()
            cursors.append(cur)
            cur.execute("select 1 x union select " + str(i))
        con.close()

    @unittest.skip("syntax not supported")
    def test_OnConflictRollback(self):
        con = self.con
        con.execute("create table foo(x int, unique(x) on conflict rollback)")
        con.execute("insert into foo(x) values (1)")
        try:
            con.execute("insert into foo(x) values (1)")
        except monetdbe.DatabaseError:
            pass
        con.execute("insert into foo(x) values (2)")
        try:
            con.commit()
        except monetdbe.OperationalError:
            self.fail("pymonetdbe knew nothing about the implicit ROLLBACK")

    def test_WorkaroundForBuggymonetdbeTransferBindings(self):
        """
        pymonetdbe would crash with older monetdbe versions unless
        a workaround is implemented.
        """
        self.con.execute("create table foo(bar int)")
        self.con.execute("drop table foo")
        self.con.execute("create table foo(bar int)")

    def test_EmptyStatement(self):
        """
        pymonetdbe used to segfault with monetdbe versions 3.5.x. These return NULL
        for "no-operation" statements
        """
        with self.assertRaises(monetdbe.OperationalError):
            self.con.execute("")

    @unittest.skip("not supported (yet)")
    def test_TypeMapUsage(self):
        """
        pymonetdbe until 2.4.1 did not rebuild the row_cast_map when recompiling
        a statement. This test exhibits the problem.
        """
        SELECT = "select * from foo"

        con = self.con
        con.close()
        con = monetdbe.connect(":memory:", detect_types=monetdbe.PARSE_DECLTYPES)
        con.execute("create table foo(bar timestamp)")
        con.execute("insert into foo(bar) values (?)", (datetime.datetime.now(),))
        con.execute(SELECT)
        con.execute("drop table foo")
        con.execute("create table foo(bar integer)")
        con.execute("insert into foo(bar) values (5)")
        con.execute(SELECT)
        self.con = con

    @unittest.skip("todo/note (gijs): disable this for now since the monetdbe engine sees this as valid")
    def test_ErrorMsgDecodeError(self):
        # When porting the module to Python 3.0, the error message about
        # decoding errors disappeared. This verifies they're back again.
        with self.assertRaises(monetdbe.OperationalError) as cm:
            self.con.execute("select 'xxx' || ? || 'yyy' colname",
                             (bytes(bytearray([250])),)).fetchone()

    def test_RegisterAdapter(self):
        """
        See issue 3312.
        """
        self.assertRaises(TypeError, monetdbe.register_adapter, {}, None)

    @unittest.skip("not supported (yet)")
    def test_SetIsolationLevel(self):
        # See issue 27881.
        class CustomStr(str):
            def upper(self):
                return None

            def __del__(self):
                con.isolation_level = ""

        con = monetdbe.connect(":memory:")
        con.isolation_level = None
        for level in "", "DEFERRED", "IMMEDIATE", "EXCLUSIVE":
            with self.subTest(level=level):
                con.isolation_level = level
                con.isolation_level = level.lower()
                con.isolation_level = level.capitalize()
                con.isolation_level = CustomStr(level)

        # setting isolation_level failure should not alter previous state
        con.isolation_level = None
        con.isolation_level = "DEFERRED"
        pairs = [
            (1, TypeError), (b'', TypeError), ("abc", ValueError),
            ("IMMEDIATE\0EXCLUSIVE", ValueError), ("\xe9", ValueError),
        ]
        for value, exc in pairs:
            with self.subTest(level=value):
                with self.assertRaises(exc):
                    con.isolation_level = value
                self.assertEqual(con.isolation_level, "DEFERRED")

    # def test_CursorConstructorCallCheck(self):
    #    """
    #    Verifies that cursor methods check whether base class __init__ was
    #    called.
    #    """
    #    class Cursor(monetdbe.Cursor):
    #        def __init__(self, con):
    #            pass

    #    con = monetdbe.connect(":memory:")
    #    cur = Cursor(con)
    #    with self.assertRaises(monetdbe.ProgrammingError):
    #        cur.execute("select 4+5").fetchall()
    #    with self.assertRaisesRegex(monetdbe.ProgrammingError,
    #                                r'^Base Cursor\.__init__ not called\.$'):
    #        cur.close()

    def test_StrSubclass(self):
        """
        The Python 3.0 port of the module didn't cope with values of subclasses of str.
        """

        class MyStr(str):
            pass

        self.con.execute("select cast(? as varchar(3))", (MyStr("abc"),))

    def test_ConnectionConstructorCallCheck(self):
        """
        Verifies that connection methods check whether base class __init__ was
        called.
        """

        class Connection(monetdbe.Connection):
            def __init__(self, name):
                pass

        con = Connection(":memory:")
        with self.assertRaises(monetdbe.ProgrammingError):
            cur = con.cursor()

    def test_CursorRegistration(self):
        """
        Verifies that subclassed cursor classes are correctly registered with
        the connection object, too.  (fetch-across-rollback problem)
        """

        class Connection(monetdbe.Connection):
            def cursor(self, *args, **kwargs):
                return Cursor(self)

        class Cursor(monetdbe.Cursor):
            def __init__(self, con):
                monetdbe.Cursor.__init__(self, con)

        con = Connection(":memory:")
        cur = con.cursor()
        cur.execute("create table foo(x int)")
        cur.executemany("insert into foo(x) values (?)", [(3,), (4,), (5,)])
        cur.execute("select x from foo")
        con.rollback()
        with self.assertRaises(monetdbe.InterfaceError):
            cur.fetchall()

    @unittest.skip("we don't support isolation_level yet")
    def test_AutoCommit(self):
        """
        Verifies that creating a connection in autocommit mode works.
        2.5.3 introduced a regression so that these could no longer
        be created.
        """
        con = monetdbe.connect(":memory:", isolation_level=None)

    # def test_PragmaAutocommit(self):
    #    """
    #    Verifies that running a PRAGMA statement that does an autocommit does
    #    work. This did not work in 2.5.3/2.5.4.
    #    """
    #    cur = self.con.cursor()
    #    cur.execute("create table foo(bar)")
    #    cur.execute("insert into foo(bar) values (5)")

    #    cur.execute("pragma page_size")
    #    row = cur.fetchone()

    def test_ConnectionCall(self):
        """
        Call a connection with a non-string SQL request: check error handling
        of the statement constructor.
        """
        self.assertRaises(TypeError, self.con, 1)

    @unittest.skip("collation not implemented yet")
    def test_Collation(self):
        def collation_cb(a, b):
            return 1

        self.assertRaises(monetdbe.ProgrammingError, self.con.create_collation,
                          # Lone surrogate cannot be encoded to the default encoding (utf8)
                          "\uDC80", collation_cb)

    def test_RecursiveCursorUse(self):
        # note: modified test slighty since we actually just handle this fine.
        con = self.con

        cur = con.cursor()
        cur.execute("create table a (bar int)")
        cur.execute("create table b (baz int)")

        def foo():
            cur.execute("insert into a (bar) values (?)", (1,))
            yield 1

        cur.executemany("insert into b (baz) values (?)",
                        ((i,) for i in foo()))

    @unittest.skip("detect types not supported (yet)")
    def test_ConvertTimestampMicrosecondPadding(self):
        """
        http://bugs.python.org/issue14720

        The microsecond parsing of convert_timestamp() should pad with zeros,
        since the microsecond string "456" actually represents "456000".
        """

        con = self.con
        con.close()
        con = monetdbe.connect(":memory:", detect_types=monetdbe.PARSE_DECLTYPES)
        cur = con.cursor()
        cur.execute("CREATE TABLE t (x TIMESTAMP)")

        # Microseconds should be 456000
        cur.execute("INSERT INTO t (x) VALUES ('2012-04-04 15:06:00.456')")

        # Microseconds should be truncated to 123456
        cur.execute("INSERT INTO t (x) VALUES ('2012-04-04 15:06:00.123456789')")

        cur.execute("SELECT * FROM t")
        values = [x[0] for x in cur.fetchall()]

        self.assertEqual(values, [
            datetime.datetime(2012, 4, 4, 15, 6, 0, 456000),
            datetime.datetime(2012, 4, 4, 15, 6, 0, 123456),
        ])
        self.con = con

    def test_InvalidIsolationLevelType(self):
        # isolation level is a string, not an integer

        self.con.close()
        self.assertRaises(TypeError,
                          monetdbe.connect, ":memory:", isolation_level=123)

    @unittest.skip("We don't handle \0 cases yet")
    def test_NullCharacter(self):
        # Issue #21147
        con = monetdbe.connect(":memory:")
        self.assertRaises(ValueError, con, "\0select 1")
        self.assertRaises(ValueError, con, "select 1\0")
        cur = con.cursor()
        self.assertRaises(ValueError, cur.execute, " \0select 2")
        self.assertRaises(ValueError, cur.execute, "select 2\0")

    @unittest.skip("isolation_level not implemented yet")
    def test_CommitCursorReset(self):
        """
        Connection.commit() did reset cursors, which made monetdbe
        to return rows multiple times when fetched from cursors
        after commit. See issues 10513 and 23129 for details.
        """
        con = monetdbe.connect(":memory:")
        con.executescript("""
        create table t(c int);
        create table t2(c int);
        insert into t values(0);
        insert into t values(1);
        insert into t values(2);
        """)

        self.assertEqual(con.isolation_level, "")

        counter = 0
        for i, row in enumerate(con.execute("select c from t")):
            with self.subTest(i=i, row=row):
                con.execute("insert into t2(c) values (?)", (i,))
                con.commit()
                if counter == 0:
                    self.assertEqual(row[0], 0)
                elif counter == 1:
                    self.assertEqual(row[0], 1)
                elif counter == 2:
                    self.assertEqual(row[0], 2)
                counter += 1
        self.assertEqual(counter, 3, "should have returned exactly three rows")

    def test_Bpo31770(self):
        """
        The interpreter shouldn't crash in case Cursor.__init__() is called
        more than once.
        """

        def callback(*args):
            pass

        con = monetdbe.connect(":memory:")
        cur = monetdbe.Cursor(con)
        ref = weakref.ref(cur, callback)
        cur.__init__(con)
        del cur
        # The interpreter shouldn't crash when ref is collected.
        del ref
        gc.collect()
        con.close()

    @unittest.skip("not supported (yet)")
    def test_DelIsolation_levelSegfault(self):
        with self.assertRaises(AttributeError):
            del self.con.isolation_level

    @unittest.skip("set_trace_callback not implemented yet")
    def test_Bpo37347(self):
        class Printer:
            def log(self, *args):
                return monetdbe.monetdbe_OK

        for method in [self.con.set_trace_callback,
                       functools.partial(self.con.set_progress_handler, n=1),
                       self.con.set_authorizer]:
            printer_instance = Printer()
            method(printer_instance.log)
            method(printer_instance.log)
            self.con.execute("select 1")  # trigger seg fault
            method(None)


class TestMonetDBeRegressions(unittest.TestCase):
    def test_crash_on_url(self):
        with self.assertRaises(Exception):
            monetdbe.connect("monetdb://localhost:5000/sf1?user=monetdb&password=monetdb")

    def test_multiple_memory_db_issue60(self):
        q = "create table test(i int)"
        m = monetdbe.connect()
        c = m.cursor()
        c.execute(q)
        m.close()
        m = monetdbe.connect()
        c = m.cursor()
        c.execute(q)
        m.close()

    def test_proper_error_on_empty_query_issue63(self):
        conn = monetdbe.connect(':memory:')
        with self.assertRaises(monetdbe.OperationalError):
            conn.execute("")

        with self.assertRaises(monetdbe.OperationalError):
            conn.execute(";")
        conn.close()

    def test_real_issue83(self):
        conn = monetdbe.connect(':memory:')
        cursor = conn.cursor()
        cursor.execute('CREATE TABLE "test"("a" REAL);')

        df = pd.DataFrame({'a': [1, 2, 3, 4]}, dtype=np.float32)
        cursor.insert('test', df)

        cursor.execute('SELECT * FROM "test"')
        df_out = cursor.fetchdf()
        pd.testing.assert_frame_equal(df, df_out)
        conn.close()

    def test_relative_path(self):
        path = Path('this_folder_can_be_removed')

        def clean():
            if path.exists():
                rmtree(path, ignore_errors=True)

        clean()
        try:
            con = monetdbe.connect(str(path))
            x = con.execute('select * from sys.tables').fetchall()
            con.close()
        except Exception:
            clean()
            raise
        else:
            clean()

    @unittest.skip("issue #84")
    def test_copy_into_issue84(self):
        con = monetdbe.connect()
        cur = con.execute("""
        CREATE TABLE test (
            i int,
            s string,
            i2 int,
            f float)
        """)

        path = str((Path(__file__).parent / "example.csv").resolve().absolute())
        cur.execute(f"COPY  INTO test FROM '{path}' delimiters ',','\n'  best effort")
        con.close()

    @unittest.skip("Disabled since takes quite long")
    def test_crash_loop(self):
        for i in range(1000):
            cx = monetdbe.connect(":memory:")
            cu = cx.cursor()
            cu.execute("create table test(id integer auto_increment primary key, name text)")
            cu.execute("insert into test(name) values (?)", ("foo",))
            cx.close()

    def test_issue127(self):
        conn = monetdbe.connect(':memory:')
        cur = conn.cursor()
        res = cur.execute("create table tmp(i integer, s string)")
        res = cur.execute("insert into tmp values(123, 'hello''world'':\n ERROR');")
        rows = res.fetchall()
        print(rows)
        conn.close()

    @unittest.skip("Disabled until issue #118 is solved")
    def test_issue118_explain(self):
        conn = monetdbe.connect(':memory:')
        cur = conn.cursor()
        res = cur.execute('explain select 1')
        tbl = res.fetchall()
        print(tbl)
        conn.close()

    @unittest.skipIf(OCT2020, "This issue was not fixed on Oct2020")
    def test_issue_136_bigint_result(self):
        con = monetdbe.connect(':memory:')
        cur = con.execute("""
        CREATE TABLE test (
            col1 INT NOT NULL,
            col2 INT NOT NULL,
            col3 INT NOT NULL,
            col4 INT NOT NULL
        )
        """)
        np.random.seed(0)
        df = pd.DataFrame(
            np.random.randint(low=np.iinfo(np.int32).min, high=np.iinfo(np.int32).max, dtype=np.int32, size=(100, 4)),
            columns=('col1', 'col2', 'col3', 'col4'))
        cur.insert('test', df)
        a = np.array([-781377998], dtype=np.int32)
        cur.insert('test', {'col1': a, 'col2': a, 'col3': a, 'col4': a})
        res5 = cur.execute('select * from test where col1 = -781377998').fetchall()
        assert (len(res5) == 1)
        assert (len(res5[0]) == 4)
        con.close()



# -*- coding: iso-8859-1 -*-
# monetdbe/test/test_dbapi.py: tests for DB-API compliance
#
# Copyright (C) 2004-2010 Gerhard H�ring <gh@ghaering.de>
#
# This file is part of pymonetdbe.
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

import threading
import unittest
from shutil import rmtree
from tempfile import NamedTemporaryFile

import monetdbe

TESTFN = NamedTemporaryFile().name


class ModuleTests(unittest.TestCase):
    def test_APILevel(self):
        self.assertEqual(monetdbe.apilevel, "2.0",
                         "apilevel is %s, should be 2.0" % monetdbe.apilevel)

    def test_ThreadSafety(self):
        self.assertEqual(monetdbe.threadsafety, 1,
                         "threadsafety is %d, should be 1" % monetdbe.threadsafety)

    def test_ParamStyle(self):
        self.assertEqual(monetdbe.paramstyle, "qmark",
                         "paramstyle is '%s', should be 'qmark'" %
                         monetdbe.paramstyle)

    def test_Warning(self):
        self.assertTrue(issubclass(monetdbe.Warning, Exception),
                        "Warning is not a subclass of Exception")

    def test_Error(self):
        self.assertTrue(issubclass(monetdbe.Error, Exception),
                        "Error is not a subclass of Exception")

    def test_InterfaceError(self):
        self.assertTrue(issubclass(monetdbe.InterfaceError, monetdbe.Error),
                        "InterfaceError is not a subclass of Error")

    def test_DatabaseError(self):
        self.assertTrue(issubclass(monetdbe.DatabaseError, monetdbe.Error),
                        "DatabaseError is not a subclass of Error")

    def test_DataError(self):
        self.assertTrue(issubclass(monetdbe.DataError, monetdbe.DatabaseError),
                        "DataError is not a subclass of DatabaseError")

    def test_OperationalError(self):
        self.assertTrue(issubclass(monetdbe.OperationalError, monetdbe.DatabaseError),
                        "OperationalError is not a subclass of DatabaseError")

    def test_IntegrityError(self):
        self.assertTrue(issubclass(monetdbe.IntegrityError, monetdbe.DatabaseError),
                        "IntegrityError is not a subclass of DatabaseError")

    def test_InternalError(self):
        self.assertTrue(issubclass(monetdbe.InternalError, monetdbe.DatabaseError),
                        "InternalError is not a subclass of DatabaseError")

    def test_ProgrammingError(self):
        self.assertTrue(issubclass(monetdbe.ProgrammingError, monetdbe.DatabaseError),
                        "ProgrammingError is not a subclass of DatabaseError")

    def test_NotSupportedError(self):
        self.assertTrue(issubclass(monetdbe.NotSupportedError,
                                   monetdbe.DatabaseError),
                        "NotSupportedError is not a subclass of DatabaseError")


class ConnectionTests(unittest.TestCase):

    def setUp(self):
        self.cx = monetdbe.connect(":memory:")
        cu = self.cx.cursor()
        # todo/note (Gijs): made ID auto_increment
        cu.execute("create table test(id integer auto_increment primary key, name text)")
        cu.execute("insert into test(name) values (?)", ("foo",))

    def tearDown(self):
        self.cx.close()

    def test_Commit(self):
        self.cx.commit()

    def test_CommitAfterNoChanges(self):
        """
        A commit should also work when no changes were made to the database.
        """
        self.cx.commit()
        self.cx.commit()

    def test_Rollback(self):
        self.cx.rollback()

    def test_RollbackAfterNoChanges(self):
        """
        A rollback should also work when no changes were made to the database.
        """
        self.cx.rollback()
        self.cx.rollback()

    def test_Cursor(self):
        cu = self.cx.cursor()

    def test_Close(self):
        self.cx.close()

    def test_Exceptions(self):
        # Optional DB-API extension.
        self.assertEqual(self.cx.Warning, monetdbe.Warning)
        self.assertEqual(self.cx.Error, monetdbe.Error)
        self.assertEqual(self.cx.InterfaceError, monetdbe.InterfaceError)
        self.assertEqual(self.cx.DatabaseError, monetdbe.DatabaseError)
        self.assertEqual(self.cx.DataError, monetdbe.DataError)
        self.assertEqual(self.cx.OperationalError, monetdbe.OperationalError)
        self.assertEqual(self.cx.IntegrityError, monetdbe.IntegrityError)
        self.assertEqual(self.cx.InternalError, monetdbe.InternalError)
        self.assertEqual(self.cx.ProgrammingError, monetdbe.ProgrammingError)
        self.assertEqual(self.cx.NotSupportedError, monetdbe.NotSupportedError)

    def test_InTransactionRO(self):
        with self.assertRaises(AttributeError):
            self.cx.in_transaction = True

    @unittest.skip("TODO: Not yet implemented, see issue #22")
    def test_OpenUri(self):
        self.addCleanup(rmtree, TESTFN)
        with monetdbe.connect(TESTFN) as cx:
            cx.execute('create table test(id integer)')
        with monetdbe.connect('file:' + TESTFN, uri=True) as cx:
            cx.execute('insert into test(id) values(0)')
        with monetdbe.connect('file:' + TESTFN + '?mode=ro', uri=True) as cx:
            with self.assertRaises(monetdbe.OperationalError):
                cx.execute('insert into test(id) values(1)')


class VariousConnectionTests(unittest.TestCase):
    def test_OpenWithPathLikeObject(self):
        """ Checks that we can successfully connect to a database using an object that
            is PathLike, i.e. has __fspath__(). """
        self.addCleanup(lambda x: rmtree(x, ignore_errors=True), TESTFN)

        class Path:
            def __fspath__(self):
                return TESTFN

        path = Path()
        with monetdbe.connect(path) as cx:
            cx.execute('create table test(id integer)')

    def test_InTransaction(self):
        # Can't use db from setUp because we want to test initial state.
        with monetdbe.connect(":memory:") as cx:
            cu = cx.cursor()
            self.assertEqual(cx.in_transaction, False)
            cu.execute("create table transactiontest(id integer auto_increment primary key, name text)")
            self.assertEqual(cx.in_transaction, True)
            cu.execute("insert into transactiontest(name) values (?)", ("foo",))
            self.assertEqual(cx.in_transaction, True)
            cu.execute("select name from transactiontest where name=?", ["foo"])
            row = cu.fetchone()
            self.assertEqual(cx.in_transaction, True)
            cx.commit()
            self.assertEqual(cx.in_transaction, False)
            cu.execute("select name from transactiontest where name=?", ["foo"])
            row = cu.fetchone()
            self.assertEqual(cx.in_transaction, True)

    def test_FailedOpen(self):
        import os
        YOU_CANNOT_OPEN_THIS = os.path.join("foo", "bar", "bla", "23534", "mydb.db")
        with self.assertRaises(monetdbe.OperationalError):
            con = monetdbe.connect(YOU_CANNOT_OPEN_THIS)

    def test_empty_query(self):
        with monetdbe.connect(':memory:') as conn:
            with self.assertRaises(monetdbe.OperationalError):
                conn.execute("")

            with self.assertRaises(monetdbe.OperationalError):
                conn.execute(";")


def prepare_cursor(cx):
    cu = cx.cursor()

    # todo/note (Gijs): changed income type from number to float and made ID auto_increment
    cu.execute(
        "create table test(id integer auto_increment primary key, name text, "
        "income float, unique_test text unique)"
    )
    cu.execute("insert into test(name) values (?)", ("foo",))
    return cu


class CursorTests(unittest.TestCase):
    def setUp(self):
        self.cx = monetdbe.connect(":memory:")
        self.cu = self.cx.cursor()

        # todo/note (Gijs): changed income type from number to float and made ID auto_increment
        self.cu.execute(
            "create table test(id integer auto_increment primary key, name text, "
            "income float, unique_test text unique)"
        )
        self.cu.execute("insert into test(name) values (?)", ("foo",))

    def tearDown(self):
        self.cu.close()
        self.cx.close()

    def test_ExecuteNoArgs(self):
        self.cu.execute("delete from test")

    def test_ExecuteIllegalSql(self):
        with self.assertRaises(monetdbe.OperationalError):
            self.cu.execute("select asdf")

    @unittest.skip("todo/note (gijs): disable this for now since the monetdbe engine sees this as valid")
    def test_ExecuteTooMuchSql(self):
        with self.assertRaises(monetdbe.OperationalError):
            self.cu.execute("select 5+4; select 4+5")

    def test_ExecuteTooMuchSql2(self):
        self.cu.execute("select 5+4; -- foo bar")

    def test_ExecuteTooMuchSql3(self):
        self.cu.execute("""
            select 5+4;

            /*
            foo
            */
            """)

    def test_empty_query(self):
        with self.assertRaises(monetdbe.OperationalError):
            self.cx.execute("")

        with self.assertRaises(monetdbe.OperationalError):
            self.cx.execute(";")

    def test_ExecuteWrongSqlArg(self):
        with self.assertRaises(monetdbe.OperationalError):
            self.cu.execute(42)

    def test_ExecuteArgInt(self):
        self.cu.execute("insert into test(id) values (?)", (42,))

    def test_ExecuteArgFloat(self):
        self.cu.execute("insert into test(income) values (?)", (2500.32,))

    def test_ExecuteArgString(self):
        self.cu.execute("insert into test(name) values (?)", ("Hugo",))

    @unittest.skip("todo: not yet supported, see issue #21")
    def test_ExecuteArgStringWithZeroByte(self):
        self.cu.execute("insert into test(name) values (?)", ("Hu\x00go",))

        self.cu.execute("select name from test where id=?", (self.cu.lastrowid,))
        row = self.cu.fetchone()
        self.assertEqual(row[0], "Hu\x00go")

    def test_ExecuteNonIterable(self):
        with self.assertRaises(ValueError) as cm:
            self.cu.execute("insert into test(id) values (?)", 42)
        self.assertEqual(str(cm.exception), "parameters '42' type '<class 'int'>' not supported")

    def test_ExecuteWrongNoOfArgs1(self):
        # too many parameters
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.execute("insert into test(id) values (?)", (17, "Egon"))

    def test_ExecuteWrongNoOfArgs2(self):
        # too little parameters
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.execute("insert into test(id) values (?)")

    def test_ExecuteWrongNoOfArgs3(self):
        # no parameters, parameters are needed
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.execute("insert into test(id) values (?)")

    def test_ExecuteParamList(self):
        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("select name from test where name=?", ["foo"])
        row = self.cu.fetchone()
        self.assertEqual(row[0], "foo")

    def test_ExecuteParamSequence(self):
        class L:
            def __len__(self):
                return 1

            def __getitem__(self, x):
                assert x == 0
                return "foo"

        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("select name from test where name=?", L())
        row = self.cu.fetchone()
        self.assertEqual(row[0], "foo")

    def test_ExecuteDictMapping(self):
        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("select name from test where name=:name", {"name": "foo"})
        row = self.cu.fetchone()
        self.assertEqual(row[0], "foo")

    def test_ExecuteDictMapping_Mapping(self):
        class D(dict):
            def __missing__(self, key):
                return "foo"

        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("select name from test where name=:name", D(), paramstyle="named")
        row = self.cu.fetchone()
        self.assertEqual(row[0], "foo")

    def test_ExecuteDictMappingTooLittleArgs(self):
        self.cu.execute("insert into test(name) values ('foo')")
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.execute("select name from test where name=:name and id=:id", {"name": "foo"})

    def test_ExecuteDictMappingNoArgs(self):
        self.cu.execute("insert into test(name) values ('foo')")
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.execute("select name from test where name=:name", paramstyle="named")

    def test_ExecuteDictMappingUnnamed(self):
        self.cu.execute("insert into test(name) values ('foo')")
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.execute("select name from test where name=?", {"name": "foo"})

    def test_Close(self):
        self.cu.close()

    def test_RowcountExecute(self):
        self.cu.execute("delete from test")
        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("update test set name='bar'")
        self.assertEqual(self.cu.rowcount, 2)

    @unittest.skip("we skip this test, since we actually *do* know the rowcount")
    def test_RowcountSelect(self):
        """
        pymonetdbe does not know the rowcount of SELECT statements, because we
        don't fetch all rows after executing the select statement. The rowcount
        has thus to be -1.
        """
        self.cu.execute("select 5 union select 6")
        self.assertEqual(self.cu.rowcount, -1)

    def test_RowcountExecutemany(self):
        self.cu.execute("delete from test")
        self.cu.executemany("insert into test(name) values (?)", [(1,), (2,), (3,)])
        self.assertEqual(self.cu.rowcount, 3)

    def test_TotalChanges(self):
        self.cu.execute("insert into test(name) values ('foo')")
        self.cu.execute("insert into test(name) values ('foo')")
        self.assertLess(2, self.cx.total_changes, msg='total changes reported wrong value')

    # Checks for executemany:
    # Sequences are required by the DB-API, iterators
    # enhancements in pymonetdbe.

    def test_ExecuteManySequence(self):
        self.cu.executemany("insert into test(income) values (?)", [(x,) for x in range(100, 110)])

    def test_ExecuteManyIterator(self):
        class MyIter:
            def __init__(self):
                self.value = 5

            def __next__(self):
                if self.value == 10:
                    raise StopIteration
                else:
                    self.value += 1
                    return (self.value,)

        self.cu.executemany("insert into test(income) values (?)", MyIter())

    def test_ExecuteManyGenerator(self):
        def mygen():
            for i in range(5):
                yield (i,)

        self.cu.executemany("insert into test(income) values (?)", mygen())

    def test_ExecuteManyWrongSqlArg(self):
        with self.assertRaises(TypeError):
            self.cu.executemany(42, [(3,)])

    def test_ExecuteManySelect(self):
        with self.assertRaises(monetdbe.ProgrammingError):
            self.cu.executemany("select ?", [(3,)])

    def test_ExecuteManyNotIterable(self):
        with self.assertRaises(TypeError):
            self.cu.executemany("insert into test(income) values (?)", 42)

    def test_FetchIter(self):
        # Optional DB-API extension.
        self.cu.execute("delete from test")
        self.cu.execute("insert into test(id) values (?)", (5,))
        self.cu.execute("insert into test(id) values (?)", (6,))
        self.cu.execute("select id from test order by id")
        lst = []
        for row in self.cu:
            lst.append(row[0])
        self.assertEqual(lst[0], 5)
        self.assertEqual(lst[1], 6)

    def test_Fetchone(self):
        self.cu.execute("select name from test")
        row = self.cu.fetchone()
        self.assertEqual(row[0], "foo")
        row = self.cu.fetchone()
        self.assertEqual(row, None)

    @unittest.skip(
        """
        monetdbe returns the parameter description table for PREPARED statements.
        Not really sure if it is a bad thing that the cursor can fetch that information""")
    def test_FetchoneNoStatement(self):
        cur = self.cx.cursor()
        row = cur.fetchone()
        self.assertEqual(row, None)

    def test_ArraySize(self):
        # must default ot 1
        self.assertEqual(self.cu.arraysize, 1)

        # now set to 2
        self.cu.arraysize = 2

        # now make the query return 3 rows
        self.cu.execute("delete from test")
        self.cu.execute("insert into test(name) values ('A')")
        self.cu.execute("insert into test(name) values ('B')")
        self.cu.execute("insert into test(name) values ('C')")
        self.cu.execute("select name from test")
        res = self.cu.fetchmany()

        self.assertEqual(len(res), 2)

    def test_Fetchmany(self):
        self.cu.execute("select name from test")
        res = self.cu.fetchmany(100)
        self.assertEqual(len(res), 1)
        res = self.cu.fetchmany(100)
        self.assertEqual(res, [])

    def test_FetchmanyKwArg(self):
        """Checks if fetchmany works with keyword arguments"""
        self.cu.execute("select name from test")
        res = self.cu.fetchmany(size=100)
        self.assertEqual(len(res), 1)

    def test_Fetchall(self):
        self.cu.execute("select name from test")
        res = self.cu.fetchall()
        self.assertEqual(len(res), 1)
        res = self.cu.fetchall()
        self.assertEqual(res, [])

    def test_Setinputsizes(self):
        self.cu.setinputsizes([3, 4, 5])

    def test_Setoutputsize(self):
        self.cu.setoutputsize(5, 0)

    def test_SetoutputsizeNoColumn(self):
        self.cu.setoutputsize(42)

    def test_CursorConnection(self):
        # Optional DB-API extension.
        self.assertEqual(self.cu.connection, self.cx)

    def test_WrongCursorCallable(self):
        with self.assertRaises(TypeError):
            def f():
                ...

            cur = self.cx.cursor(f)

    def test_CursorWrongClass(self):
        class Foo:
            ...

        foo = Foo()
        with self.assertRaises(TypeError):
            cur = monetdbe.Cursor(foo)

    @unittest.skip("monetdb doesn't support INSERT OR REPLACE")
    def test_LastRowIDOnReplace(self):
        """
        INSERT OR REPLACE and REPLACE INTO should produce the same behavior.
        """
        sql = '{} INTO test(id, unique_test) VALUES (?, ?)'
        for statement in ('INSERT OR REPLACE', 'REPLACE'):
            with self.subTest(statement=statement):
                self.cu.execute(sql.format(statement), (1, 'foo'))
                self.assertEqual(self.cu.lastrowid, 1)

    @unittest.skip("skipping this test, we don't support 'insert or '")
    def test_LastRowIDOnIgnore(self):
        self.cu.execute(
            "insert or ignore into test(unique_test) values (?)",
            ('test',))
        self.assertEqual(self.cu.lastrowid, 2)
        self.cu.execute(
            "insert or ignore into test(unique_test) values (?)",
            ('test',))
        self.assertEqual(self.cu.lastrowid, 2)

    @unittest.skip("skipping this test, we don't support 'insert or '")
    def test_LastRowIDInsertOR(self):
        results = []
        for statement in ('FAIL', 'ABORT', 'ROLLBACK'):
            sql = 'INSERT OR {} INTO test(unique_test) VALUES (?)'
            with self.subTest(statement='INSERT OR {}'.format(statement)):
                formatted = sql.format(statement)
                self.cu.execute(formatted, (statement,))
                results.append((statement, self.cu.lastrowid))
                with self.assertRaises(monetdbe.IntegrityError):
                    self.cu.execute(sql.format(statement), (statement,))
                results.append((statement, self.cu.lastrowid))
        expected = [
            ('FAIL', 2), ('FAIL', 2),
            ('ABORT', 3), ('ABORT', 3),
            ('ROLLBACK', 4), ('ROLLBACK', 4),
        ]
        self.assertEqual(results, expected)


@unittest.skip("todo (gijs): for now we dont support or check for multi-threading")
class ThreadTests(unittest.TestCase):
    def setUp(self):
        self.con = monetdbe.connect(":memory:")
        self.cur = self.con.cursor()
        # NOTE: (gijs) replaced binary type with blob
        self.cur.execute("create table test(id integer primary key, name text, bin blob, ratio float, ts timestamp)")

    def tearDown(self):
        self.cur.close()
        self.con.close()

    def test_ConCursor(self):
        def run(con, errors):
            try:
                cur = con.cursor()
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        t = threading.Thread(target=run, kwargs={"con": self.con, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_ConCommit(self):
        def run(con, errors):
            try:
                con.commit()
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        t = threading.Thread(target=run, kwargs={"con": self.con, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_ConRollback(self):
        def run(con, errors):
            try:
                con.rollback()
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        t = threading.Thread(target=run, kwargs={"con": self.con, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_ConClose(self):
        def run(con, errors):
            try:
                con.close()
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        t = threading.Thread(target=run, kwargs={"con": self.con, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_CurImplicitBegin(self):
        def run(cur, errors):
            try:
                cur.execute("insert into test(name) values ('a')")
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        t = threading.Thread(target=run, kwargs={"cur": self.cur, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_CurClose(self):
        def run(cur, errors):
            try:
                cur.close()
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        t = threading.Thread(target=run, kwargs={"cur": self.cur, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_CurExecute(self):
        def run(cur, errors):
            try:
                cur.execute("select name from test")
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        self.cur.execute("insert into test(name) values ('a')")
        t = threading.Thread(target=run, kwargs={"cur": self.cur, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))

    def test_CurIterNext(self):
        def run(cur, errors):
            try:
                row = cur.fetchone()
                errors.append("did not raise ProgrammingError")
                return
            except monetdbe.ProgrammingError:
                return
            except Exception:
                errors.append("raised wrong exception")

        errors = []
        self.cur.execute("insert into test(name) values ('a')")
        self.cur.execute("select name from test")
        t = threading.Thread(target=run, kwargs={"cur": self.cur, "errors": errors})
        t.start()
        t.join()
        if len(errors) > 0:
            self.fail("\n".join(errors))


class ConstructorTests(unittest.TestCase):
    def test_Date(self):
        d = monetdbe.Date(2004, 10, 28)

    def test_Time(self):
        t = monetdbe.Time(12, 39, 35)

    def test_Timestamp(self):
        ts = monetdbe.Timestamp(2004, 10, 28, 12, 39, 35)

    def test_DateFromTicks(self):
        d = monetdbe.DateFromTicks(42)

    def test_TimeFromTicks(self):
        t = monetdbe.TimeFromTicks(42)

    def test_TimestampFromTicks(self):
        ts = monetdbe.TimestampFromTicks(42)

    def test_Binary(self):
        b = monetdbe.Binary(b"\0'")


from monetdbe._cffi.branch import newer_then_jul2021
if newer_then_jul2021:
    class DecimalTests(unittest.TestCase):
        def setUp(self):
            self.con = monetdbe.connect(":memory:")
            self.cur = self.con.cursor()
            # NOTE: (gijs) replaced binary type with blob
            self.cur.execute("create table test(x DECIMAL(10,5))")

        def tearDown(self):
            self.cur.close()
            self.con.close()

        def test_decimals(self):
            from decimal import Decimal
            d = Decimal('12345.12345')
            self.cur.execute("insert into test VALUES (?)", [d])
            self.cur.execute("select x from test")
            row = self.cur.fetchone()
            assert(d == row[0])


class ExtensionTests(unittest.TestCase):
    def test_ScriptStringSql(self):
        with monetdbe.connect(":memory:") as con:
            cur = con.cursor()
            cur.executescript("""
                -- bla bla
                /* a stupid comment */
                create table a(i int);
                insert into a(i) values (5);
                """)
            cur.execute("select i from a")
            res = cur.fetchone()[0]
            self.assertEqual(res, 5)

    def test_ScriptSyntaxError(self):
        with monetdbe.connect(":memory:") as con:
            cur = con.cursor()
            with self.assertRaises(monetdbe.OperationalError):
                cur.executescript("create table test(x); asdf; create table test2(x)")

    def test_ScriptErrorNormal(self):
        with monetdbe.connect(":memory:") as con:
            cur = con.cursor()
            with self.assertRaises(monetdbe.OperationalError):
                cur.executescript("create table test(sadfsadfdsa); select foo from hurz;")

    def test_CursorExecutescriptAsBytes(self):
        with monetdbe.connect(":memory:") as con:
            cur = con.cursor()
            with self.assertRaises(ValueError) as cm:
                cur.executescript(b"create table test(foo); insert into test(foo) values (5);")
            self.assertEqual(str(cm.exception), 'script argument must be unicode.')

    def test_ConnectionExecute(self):
        with monetdbe.connect(":memory:") as con:
            result = con.execute("select 5").fetchone()[0]
            self.assertEqual(result, 5, "Basic test of Connection.execute")

    def test_ConnectionExecutemany(self):
        with monetdbe.connect(":memory:") as con:
            # NOTE: (gijs) added type int, required for MonetDB
            con.execute("create table test(foo int)")
            con.executemany("insert into test(foo) values (?)", [(3,), (4,)])
            result = con.execute("select foo from test order by foo").fetchall()
            self.assertEqual(result[0][0], 3, "Basic test of Connection.executemany")
            self.assertEqual(result[1][0], 4, "Basic test of Connection.executemany")

    def test_ConnectionExecutescript(self):
        with monetdbe.connect(":memory:") as con:
            # NOTE: (gijs) added type int, required for MonetDB
            con.executescript("create table test(foo int); insert into test(foo) values (5);")
            result = con.execute("select foo from test").fetchone()[0]
            self.assertEqual(result, 5, "Basic test of Connection.executescript")


class ClosedConTests(unittest.TestCase):
    def test_ClosedConCursor(self):
        with monetdbe.connect(":memory:") as con:
            con.close()
            with self.assertRaises(monetdbe.ProgrammingError):
                cur = con.cursor()

    def test_ClosedConCommit(self):
        with monetdbe.connect(":memory:") as con:
            con.close()
            with self.assertRaises(monetdbe.ProgrammingError):
                con.commit()

    def test_ClosedConRollback(self):
        with monetdbe.connect(":memory:") as con:
            con.close()
            with self.assertRaises(monetdbe.ProgrammingError):
                con.rollback()

    def test_ClosedCurExecute(self):
        with monetdbe.connect(":memory:") as con:
            cur = con.cursor()
            con.close()
            with self.assertRaises(monetdbe.ProgrammingError):
                cur.execute("select 4")

    def test_ClosedCreateFunction(self):
        con = monetdbe.connect(":memory:")
        con.close()

        def f(x):
            return 17

        with self.assertRaises(monetdbe.ProgrammingError):
            con.create_function("foo", 1, f)

    def test_ClosedCreateAggregate(self):
        con = monetdbe.connect(":memory:")
        con.close()

        class Agg:
            def __init__(self):
                pass

            def step(self, x):
                pass

            def finalize(self):
                return 17

        with self.assertRaises(monetdbe.ProgrammingError):
            con.create_aggregate("foo", 1, Agg)

    def test_ClosedSetAuthorizer(self):
        con = monetdbe.connect(":memory:")
        con.close()

        def authorizer(*args):
            return monetdbe.DENY

        with self.assertRaises(monetdbe.ProgrammingError):
            con.set_authorizer(authorizer)

    def test_ClosedSetProgressCallback(self):
        con = monetdbe.connect(":memory:")
        con.close()

        def progress():
            pass

        with self.assertRaises(monetdbe.ProgrammingError):
            con.set_progress_handler(progress, 100)

    def test_ClosedCall(self):
        con = monetdbe.connect(":memory:")
        con.close()
        with self.assertRaises(monetdbe.ProgrammingError):
            con()


class ClosedCurTests(unittest.TestCase):
    def test_Closed(self):
        with monetdbe.connect(":memory:") as con:
            cur = con.cursor()
            cur.close()

            for method_name in ("execute", "executemany", "executescript", "fetchall", "fetchmany", "fetchone"):
                if method_name in ("execute", "executescript"):
                    params = ("select 4 union select 5",)
                elif method_name == "executemany":
                    params = ("insert into foo(bar) values (?)", [(3,), (4,)])
                else:
                    params = []

                with self.assertRaises(monetdbe.ProgrammingError):
                    method = getattr(cur, method_name)
                    method(*params)


@unittest.skip("We don't support INSERT OR syntax")
class monetdbeOnConflictTests(unittest.TestCase):
    """
    Tests for monetdbe's "insert on conflict" feature.

    See https://www.monetdbe.org/lang_conflict.html for details.
    """

    def setUp(self):
        self.cx = monetdbe.connect(":memory:")
        self.cu = self.cx.cursor()
        self.cu.execute("""
          CREATE TABLE test(
            id INTEGER PRIMARY KEY auto_increment, name TEXT, unique_name TEXT UNIQUE
          );
        """)  # NOTE: (gijs) add auto_increment

    def tearDown(self):
        self.cu.close()
        self.cx.close()

    def test_OnConflictRollbackWithExplicitTransaction(self):
        self.cx.isolation_level = None  # autocommit mode
        self.cu = self.cx.cursor()
        # Start an explicit transaction.
        self.cu.execute("BEGIN")
        self.cu.execute("INSERT INTO test(name) VALUES ('abort_test')")
        self.cu.execute("INSERT OR ROLLBACK INTO test(unique_name) VALUES ('foo')")
        with self.assertRaises(monetdbe.IntegrityError):
            self.cu.execute("INSERT OR ROLLBACK INTO test(unique_name) VALUES ('foo')")
        # Use connection to commit.
        self.cx.commit()
        self.cu.execute("SELECT name, unique_name from test")
        # Transaction should have rolled back and nothing should be in table.
        self.assertEqual(self.cu.fetchall(), [])

    def test_OnConflictAbortRaisesWithExplicitTransactions(self):
        # Abort cancels the current sql statement but doesn't change anything
        # about the current transaction.
        self.cx.isolation_level = None  # autocommit mode
        self.cu = self.cx.cursor()
        # Start an explicit transaction.
        self.cu.execute("BEGIN TRANSACTION")  # NOTE: (gijs) added TRANSACTION
        self.cu.execute("INSERT INTO test(name) VALUES ('abort_test')")
        self.cu.execute("INSERT OR ABORT INTO test(unique_name) VALUES ('foo')")
        with self.assertRaises(monetdbe.IntegrityError):
            self.cu.execute("INSERT OR ABORT INTO test(unique_name) VALUES ('foo')")
        self.cx.commit()
        self.cu.execute("SELECT name, unique_name FROM test")
        # Expect the first two inserts to work, third to do nothing.
        self.assertEqual(self.cu.fetchall(), [('abort_test', None), (None, 'foo',)])

    def test_OnConflictRollbackWithoutTransaction(self):
        # Start of implicit transaction
        self.cu.execute("INSERT INTO test(name) VALUES ('abort_test')")
        self.cu.execute("INSERT OR ROLLBACK INTO test(unique_name) VALUES ('foo')")
        with self.assertRaises(monetdbe.IntegrityError):
            self.cu.execute("INSERT OR ROLLBACK INTO test(unique_name) VALUES ('foo')")
        self.cu.execute("SELECT name, unique_name FROM test")
        # Implicit transaction is rolled back on error.
        self.assertEqual(self.cu.fetchall(), [])

    def test_OnConflictAbortRaisesWithoutTransactions(self):
        # Abort cancels the current sql statement but doesn't change anything
        # about the current transaction.
        self.cu.execute("INSERT INTO test(name) VALUES ('abort_test')")
        self.cu.execute("INSERT OR ABORT INTO test(unique_name) VALUES ('foo')")
        with self.assertRaises(monetdbe.IntegrityError):
            self.cu.execute("INSERT OR ABORT INTO test(unique_name) VALUES ('foo')")
        # Make sure all other values were inserted.
        self.cu.execute("SELECT name, unique_name FROM test")
        self.assertEqual(self.cu.fetchall(), [('abort_test', None), (None, 'foo',)])

    def test_OnConflictFail(self):
        self.cu.execute("INSERT OR FAIL INTO test(unique_name) VALUES ('foo')")
        with self.assertRaises(monetdbe.IntegrityError):
            self.cu.execute("INSERT OR FAIL INTO test(unique_name) VALUES ('foo')")
        self.assertEqual(self.cu.fetchall(), [])

    def test_OnConflictIgnore(self):
        self.cu.execute("INSERT OR IGNORE INTO test(unique_name) VALUES ('foo')")
        # Nothing should happen.
        self.cu.execute("INSERT OR IGNORE INTO test(unique_name) VALUES ('foo')")
        self.cu.execute("SELECT unique_name FROM test")
        self.assertEqual(self.cu.fetchall(), [('foo',)])

    def test_OnConflictReplace(self):
        self.cu.execute("INSERT OR REPLACE INTO test(name, unique_name) VALUES ('Data!', 'foo')")
        # There shouldn't be an IntegrityError exception.
        self.cu.execute("INSERT OR REPLACE INTO test(name, unique_name) VALUES ('Very different data!', 'foo')")
        self.cu.execute("SELECT name, unique_name FROM test")
        self.assertEqual(self.cu.fetchall(), [('Very different data!', 'foo')])


def suite():
    module_suite = unittest.makeSuite(ModuleTests, "Check")
    connection_suite = unittest.makeSuite(ConnectionTests, "Check")
    cursor_suite = unittest.makeSuite(CursorTests, "Check")
    thread_suite = unittest.makeSuite(ThreadTests, "Check")
    constructor_suite = unittest.makeSuite(ConstructorTests, "Check")
    ext_suite = unittest.makeSuite(ExtensionTests, "Check")
    closed_con_suite = unittest.makeSuite(ClosedConTests, "Check")
    closed_cur_suite = unittest.makeSuite(ClosedCurTests, "Check")
    on_conflict_suite = unittest.makeSuite(monetdbeOnConflictTests, "Check")
    return unittest.TestSuite((
        module_suite, connection_suite, cursor_suite, thread_suite,
        constructor_suite, ext_suite, closed_con_suite, closed_cur_suite,
        on_conflict_suite,
    ))


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



# monetdbe/dbapi2.py: the DB-API 2.0 interface
#
# Copyright (C) 2004-2005 Gerhard Häring <gh@ghaering.de>
#
# This file is part of pymonetdbe.
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
from datetime import date, time, datetime
from time import localtime

from monetdbe.connection import Connection

OptimizedUnicode = str

paramstyle = "qmark"
threadsafety = 1
apilevel = "2.0"
Date = date
Time = time
Timestamp = datetime


def connect(*args, **kwargs) -> Connection:
    if 'factory' in kwargs:  # and 'database' in kwargs:
        factory = kwargs.pop('factory')
        return factory(*args, **kwargs)
        # return factory(database=kwargs['database'])
    return Connection(*args, **kwargs)


connect.__doc__ = Connection.__init__.__doc__


def DateFromTicks(ticks):
    return Date(*localtime(ticks)[:3])


def TimeFromTicks(ticks):
    return Time(*localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    return Timestamp(*localtime(ticks)[:6])


Binary = memoryview



from re import compile, sub, DOTALL
from string import Formatter
from typing import Dict, Optional, Union, Iterable, Any, List, Sized, Collection, Sequence, Mapping

from monetdbe.exceptions import ProgrammingError
from monetdbe.monetize import convert

# use this pattern to split a string on non-escaped semicolumns
semicolumn_split_pattern = compile(r'''((?:[^;"']|"[^"]*"|'[^']*')+)''')


def remove_quoted_substrings(query: str):
    """
    This removes all quoted substrings. Note that this likely results in a broken query.
    """
    q = sub(r"\\'", "", query)  # first remove escaped quotes
    q = sub(r"'(.*)'", "", q)  # then remove all quotes strings
    return q


def strip_split_and_clean(script: str):
    """
    This will split the query on unescaped semicolumns, cleanup every subquery from comments and remove any
    empty queries from the result.

    args:
        script: one ore more SQL queries split by a semicolum
    """

    results = []
    for q in semicolumn_split_pattern.split(script)[1::2]:
        q = sub(r'^\s*--.*\n?', '', q)
        q = sub(r'/*.*\*/', '', q, flags=DOTALL)
        q = q.strip()
        if q:
            results.append(q)

    return results


def escape(v):
    return f"'{v}'"


class DefaultFormatter(Formatter):
    """
    This makes it possible to supply a dict with a __missing__() method (like a defaultdict)
    """

    def __init__(self, d: Dict):
        super().__init__()
        self.d = d

    def get_value(self, key, *args, **kwargs):
        s = self.d[key]
        if isinstance(s, str):
            return escape(s)
        else:
            return s


parameters_type = Optional[Union[Sequence[Any], Dict[str, Any]]]


def _format_mapping(cleaned_query: str, parameters: Dict[str, Any], query: str):
    if '?' in cleaned_query:
        raise ProgrammingError("'?' in formatting with mapping as parameters")

    escaped: Dict[str, str] = {k: convert(v) for k, v in parameters.items()}

    if ':' in cleaned_query:  # qmark
        x = sub(r':(\w+)', r'{\1}', query)
    elif '%' in cleaned_query:
        return query % escaped

    if hasattr(type(parameters), '__missing__'):
        # this is something like a dict with a default value
        try:
            # mypy doesn't understand that this is a dict-like with a default __missing__ value
            return DefaultFormatter(parameters).format(x, **escaped)
        except KeyError as e:
            raise ProgrammingError(e)
    try:
        return x.format(**escaped)
    except KeyError as e:
        raise ProgrammingError(e)


def _format_iterable(cleaned_query: str, parameters: Sequence[Any], query: str):
    # we do this a bit strange to make sure the test_ExecuteParamSequence sqlite test passes
    escaped_list: List[str] = [convert(parameters[i]) for i in range(len(parameters))]

    if ':' in cleaned_query:
        x = sub(r':(\w+)', r'{\1}', query)

        # The numbering used starts at 1, while python starts 0, so we insert a bogus prefix
        prefixed = [''] + escaped_list

        return x.format(*prefixed)

    if '?' in cleaned_query:  # qmark style

        if cleaned_query.count('?') != len(escaped_list):
            raise ProgrammingError(f"Number of arguments ({len(escaped_list)}) doesn't "
                                   f"match number of '?' ({cleaned_query.count('?')})")

        return query.replace('?', '{}').format(*escaped_list)
    elif '%s' in cleaned_query:  # pyformat style
        return query % tuple(escaped_list)
    else:
        return cleaned_query


def format_query(query: str, parameters: parameters_type = None) -> str:
    if type(query) != str:
        raise TypeError

    cleaned_query = remove_quoted_substrings(query)

    if parameters is None:
        for symbol in ':?':
            if symbol in cleaned_query:
                raise ProgrammingError(f"unexpected symbol '{symbol}' in operation")
        return query

    # named, numeric or format style
    if isinstance(parameters, Dict):
        return _format_mapping(cleaned_query, parameters, query)

    # qmark or pyformat style
    elif isinstance(parameters, Sequence) or (isinstance(parameters, Sized) and hasattr(parameters, '__getitem__')):
        return _format_iterable(cleaned_query, parameters, query)

    else:
        raise ValueError(f"parameters '{parameters}' type '{type(parameters)}' not supported")





import numpy
import pandas
import pytest


class TestSimpleDBAPI:
    def test_regular_selection(self, monetdbe_cursor):
        monetdbe_cursor.execute('SELECT * FROM integers')
        result = monetdbe_cursor.fetchall()
        assert result == [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,),
                          (None,)], "Incorrect result returned"

    def test_numpy_selection(self, monetdbe_cursor):
        monetdbe_cursor.execute('SELECT * FROM integers')
        result = monetdbe_cursor.fetchnumpy()
        arr = numpy.ma.masked_array(numpy.arange(11))
        arr.mask = [False] * 10 + [True]
        numpy.testing.assert_array_equal(result['i'], arr, "Incorrect result returned")

    def test_pandas_selection(self, monetdbe_cursor):
        monetdbe_cursor.execute('SELECT * FROM integers')
        result = monetdbe_cursor.fetchdf()
        arr = numpy.ma.masked_array(numpy.arange(11))
        arr.mask = [False] * 10 + [True]
        arr = {'i': arr}
        arr = pandas.DataFrame.from_dict(arr)
        # assert str(result) == str(arr), "Incorrect result returned"
        pandas.testing.assert_frame_equal(result, arr)

    def test_numpy_creation(self, monetdbe_cursor):
        # numpyarray = {'i': numpy.arange(10), 'v': numpy.random.randint(100, size=(1, 10))}  # segfaults
        data_dict = {'i': numpy.arange(10), 'v': numpy.random.randint(100, size=10)}
        monetdbe_cursor.create('numpy_creation', data_dict)
        monetdbe_cursor.commit()

        monetdbe_cursor.execute('SELECT * FROM numpy_creation')
        result = monetdbe_cursor.fetchnumpy()

        numpy.testing.assert_array_equal(result['i'], data_dict['i'])
        numpy.testing.assert_array_equal(result['v'], data_dict['v'])

    def test_pandas_creation(self, monetdbe_cursor):
        data_dict = {'i': numpy.arange(10), 'v': numpy.random.randint(100, size=10)}
        dframe = pandas.DataFrame.from_dict(data_dict)
        monetdbe_cursor.create('dframe_creation', dframe)

        monetdbe_cursor.execute('SELECT * FROM dframe_creation')
        result = monetdbe_cursor.fetchnumpy()

        numpy.testing.assert_array_equal(result['i'], data_dict['i'])
        numpy.testing.assert_array_equal(result['v'], data_dict['v'])

    def test_numpy_insertion(self, monetdbe_cursor):
        data_dict = {'i': numpy.arange(10).astype(numpy.int32),
                     'v': numpy.random.randint(100, size=10).astype(numpy.int32)}
        monetdbe_cursor.execute("CREATE TABLE numpy_insertion (i INT, v INT)")
        monetdbe_cursor.insert('numpy_insertion', data_dict)
        monetdbe_cursor.commit()

        monetdbe_cursor.execute("SELECT * FROM numpy_insertion")
        result = monetdbe_cursor.fetchnumpy()

        numpy.testing.assert_array_equal(result['i'], data_dict['i'])
        numpy.testing.assert_array_equal(result['v'], data_dict['v'])

    def test_pandas_insertion(self, monetdbe_cursor):
        data_dict = {'i': numpy.arange(10).astype(numpy.int32),
                     'v': numpy.random.randint(100, size=10).astype(numpy.int32)}
        dframe = pandas.DataFrame.from_dict(data_dict)
        monetdbe_cursor.execute("CREATE TABLE pandas_insertion (i INT, v INT)")
        monetdbe_cursor.insert('pandas_insertion', dframe)
        monetdbe_cursor.commit()

        monetdbe_cursor.execute("SELECT * FROM pandas_insertion")
        result = monetdbe_cursor.fetchnumpy()

        numpy.testing.assert_array_equal(result['i'], data_dict['i'])
        numpy.testing.assert_array_equal(result['v'], data_dict['v'])

    def test_masked_array_insertion(self, monetdbe_cursor):
        data_dict = {'i': numpy.ma.masked_array(numpy.arange(10).astype(numpy.int32), mask=([False] * 9 + [True]))}
        monetdbe_cursor.execute("CREATE TABLE masked_array_insertion (i INT)")
        monetdbe_cursor.insert("masked_array_insertion", data_dict)
        monetdbe_cursor.commit()

        monetdbe_cursor.execute("SELECT * FROM masked_array_insertion")
        result = monetdbe_cursor.fetchnumpy()

        numpy.testing.assert_array_equal(result['i'], data_dict['i'])



"""
This module contains all python code interacting with CFFI and should not be used directly from the outside world.
"""
import logging

_logger = logging.getLogger(__name__)


def check_if_we_can_import_lowlevel():
    try:
        from monetdbe._lowlevel import lib, ffi
    except ImportError as e:
        _logger.error(e)
        _logger.error("We recommend you use the MonetDBe-Python binary wheels.")
        _logger.error("If you can't or don't want to then:")
        _logger.error("On Linux: try setting LD_LIBRARY_PATH to point to the location of libmonetdbe.so.")
        _logger.error("On OS X: try setting DYLIB_LIBRARY_PATH to point to the location of libmonetdbe.so.")
        _logger.error("On Windows: try setting PATH to point to the location of libmonetdbe.dll.")
        _logger.error("Also make sure all depending shared libraries are available.")
        raise



"""
compatibility with MonetDBLite
"""
from typing import TYPE_CHECKING
from warnings import warn
from pandas import DataFrame

from monetdbe.dbapi2 import connect
from monetdbe.connection import Connection

if TYPE_CHECKING:
    from monetdbe.cursors import Cursor  # type: ignore[attr-defined]


def make_connection(*args, **kwargs):
    warn("make_connection() is deprecated and will be removed from future versions")
    return connect(*args, **kwargs)


make_connection.__doc__ = connect.__doc__


def shutdown():
    warn("shutdown() is deprecated and will be removed from future versions")


def init(_: str):
    warn("init() is deprecated and will be removed from future versions")


def sql(query: str, client=None) -> DataFrame:
    warn("sql() is deprecated and will be removed from future versions")
    if client:
        if not isinstance(client, Connection):
            raise TypeError
        return client.execute(query).fetchdf()
    else:
        return connect().execute(query).fetchdf()


def create(table, values, schema=None, conn=None) -> 'Cursor':
    warn("create() is deprecated and will be removed from future versions")
    if not conn:
        conn = connect()
    return conn.cursor().create(table=table, values=values, schema=schema)


def insert(*args, **kwargs) -> 'Cursor':
    warn("insert() is deprecated and will be removed from future versions")
    if 'client' in kwargs:
        client = kwargs.pop('client')
        if not isinstance(client, Connection):
            raise TypeError
    else:
        client = connect()
    return client.cursor().insert(*args, **kwargs)



import re
from datetime import datetime, date, time

Binary = bytes


def strip(data):
    """ returns a python string, with chopped off quotes,
    and replaced escape characters"""
    return ''.join([w.encode('utf-8').decode('unicode_escape')
                    if '\\' in w else w
                    for w in re.split('([\000-\200]+)', data[1:-1])])


def py_bool(data):
    """ return python boolean """
    return data == "true"


def py_time(data):
    """ returns a python Time
    """
    return time(hour=data.hours, minute=data.minutes, second=data.seconds, microsecond=data.ms * 1000)


def py_date(data):
    """ Returns a python Date
    """
    return date(year=data.year, month=data.month, day=data.day)


def py_timestamp(data):
    """ Returns a python Timestamp
    """
    return datetime(year=data.date.year,
                    month=data.date.month,
                    day=data.date.day,
                    hour=data.time.hours,
                    minute=data.time.minutes,
                    second=data.time.seconds,
                    microsecond=data.time.ms * 1000)


def py_bytes(data):
    """Returns a bytes object representing the input blob."""
    return Binary(data)


def oid(data):
    """represents an object identifier
    For now we will just return the string representation just like mclient does.
    """
    return oid



import datetime
import decimal
from typing import Any, Dict, Type, Callable, Tuple, List

import numpy as np

from monetdbe.exceptions import InterfaceError


def monet_none(data: None) -> str:
    """
    returns a NULL string
    """
    return "NULL"


def monet_bool(data: bool) -> str:
    """
    returns "true" or "false"
    """
    return ["false", "true"][bool(data)]


def monet_escape(data) -> str:
    """
    returns an escaped string
    """
    data = str(data).replace("\\", "\\\\")
    data = data.replace("\'", "\\\'")
    return f"'{data}'"


def monet_identifier_escape(data) -> str:
    """
    returns an escaped identifier
    """
    data = str(data).replace("\\", "\\\\")
    data = data.replace('"', '\\"')
    return '"%s"' % str(data)


def monet_bytes(data: bytes) -> str:
    """
    converts bytes to string
    """
    return "'%s'" % data.hex()


def monet_memoryview(data: memoryview) -> str:
    return "'%s'" % data.tobytes().hex()


def monet_float(data: float) -> str:
    if data != data:  # yes this is how you can check if a float is a NaN
        return 'NULL'
    else:
        return str(data)


def monet_datetime(data: Any) -> str:
    if np.isnat(data):  # type: ignore
        return 'NULL'
    else:
        return f"'{data}'"


mapping: List[Tuple[Type, Callable]] = [
    (str, monet_escape),
    (bytes, monet_bytes),
    (memoryview, monet_memoryview),
    (int, str),
    (complex, str),
    (float, monet_float),
    (decimal.Decimal, str),
    (datetime.datetime, monet_escape),
    (datetime.time, monet_escape),
    (datetime.date, monet_escape),
    (datetime.timedelta, monet_escape),
    (bool, monet_bool),
    (type(None), monet_none),
    (np.int64, int),
    (np.int32, int),
    (np.int16, int),
    (np.int8, int),
    (np.uint64, int),
    (np.uint32, int),
    (np.uint16, int),
    (np.uint8, int),
    (np.float64, monet_float),
    (np.float32, monet_float),
    (np.datetime64, monet_datetime),  # type: ignore
    (np.ma.core.MaskedConstant, monet_none),  # type: ignore
]

mapping_dict: Dict[Type, Callable] = dict(mapping)


class PrepareProtocol:
    """
    SQLite has a PrepareProtocol class that is passed to a values __conform__ method if it exists.

    This is not well documented, and the __conform__ method seems rejected in PEP 246. Still, the SQLite
    documentation mentions this
    """
    ...


def convert(data: Any) -> str:
    """
    Return the appropriate conversion function based upon the python type.
    """
    if type(data) in mapping_dict:
        return mapping_dict[type(data)](data)
    else:
        for type_, func in mapping:
            if issubclass(type(data), type_):
                return func(data)

    if hasattr(data, '__conform__'):
        try:
            return data.__conform__(PrepareProtocol)
        except Exception as e:
            raise InterfaceError(e)

    raise InterfaceError("type %s not supported as value" % type(data))



"""
This module contains the monetdbe connection class.
"""
from collections import namedtuple
from pathlib import Path
from typing import Optional, Type, Iterable, Union, TYPE_CHECKING, Callable, Any, Iterator, Tuple, Mapping
from itertools import repeat
import numpy as np

from monetdbe import exceptions
from monetdbe.formatting import parameters_type

if TYPE_CHECKING:
    from monetdbe.row import Row
    from monetdbe.cursors import Cursor  # type: ignore[attr-defined]

Description = namedtuple('Description', (
    'name',
    'type_code',
    'display_size',
    'internal_size',
    'precision',
    'scale',
    'null_ok'
))


class Connection:
    def __init__(self,
                 database: Optional[Union[str, Path]] = None,
                 uri: bool = False,
                 timeout: int = 0,
                 detect_types: int = 0,
                 check_same_thread: bool = True,
                 autocommit: bool = False,
                 nr_threads: int = 0,
                 memorylimit: int = 0,
                 querytimeout: int = 0,
                 logging: Optional[Path] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 port: Optional[int] = None,
                 ):
        """
        Args:
            database: The path to you database. Leave empty or use the `:memory:` string to start an in-memory database.
            uri: if true, database is interpreted as a URI. This allows you to specify options.
            timeout: The session / connection timeout in seconds, 0 = no limit (default)
            detect_types:  defaults to 0 (i. e. off, no type detection), you can set it to any combination of
                           PARSE_DECLTYPES and PARSE_COLNAMES to turn type detection on.
            check_same_thread: By default, check_same_thread is True and only the creating thread may use the
                               connection. If set False, the returned connection may be shared across multiple threads.
                               When using multiple threads with the same connection writing operations should be
                               serialized by the user to avoid data corruption.
            autocommit: Enable autocommit mode
            nr_threads: to control the level of parallelism, 0 = all cores (default)
            memorylimit: to control the memory footprint allowed in :memory: mode in MB, 0 = no limit (default)
            querytimeout: The query timeout in seconds, 0 = no limit (default)
            logging: the file location for the tracer files (not used yet)
            username: used to connect to a remote server (not used yet)
            password: credentials to reach the remote server (not used yet)
            port: TCP/IP port to listen for connections (not used yet)

        """
        # import these here so we can import this file without having access to _cffi (yet)
        from monetdbe._cffi import check_if_we_can_import_lowlevel
        from monetdbe._cffi.internal import Internal
        from monetdbe._cffi.types_ import monetdbe_result

        check_if_we_can_import_lowlevel()

        if uri or port or username or password or logging:
            raise NotImplemented

        if not check_same_thread:
            raise NotImplemented

        if detect_types != 0:
            raise NotImplemented

        if not database:
            database = None
        elif database == ':memory:':  # sqlite compatibility
            database = None
        elif isinstance(database, str):
            database = Path(database).resolve()
        elif hasattr(database, '__fspath__'):  # Deal with Path like objects
            database = Path(database.__fspath__()).resolve()  # type: ignore
        else:
            raise TypeError

        self.result: Optional[monetdbe_result] = None
        self.row_factory: Optional[Type['Row']] = None
        self.text_factory: Optional[Callable[[str], Any]] = None
        self.total_changes = 0
        self.isolation_level = None
        self.consistent = True

        self._internal: Optional[Internal] = Internal(
            connection=self,
            dbdir=database,
            memorylimit=memorylimit,
            nr_threads=nr_threads,
            querytimeout=querytimeout,
            sessiontimeout=timeout
        )

        self.set_autocommit(autocommit)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __call__(self):
        raise exceptions.ProgrammingError

    def __del__(self):
        self.close()

    def _check(self):
        if not hasattr(self, '_internal') or not self._internal:
            raise exceptions.ProgrammingError("The connection has been closed")

    def get_description(self):
        # we import this late, otherwise the whole monetdbe project is unimportable
        # if we don't have access to monetdbe shared library
        from monetdbe._cffi.convert import make_string, monet_c_type_map
        from monetdbe._cffi.internal import result_fetch

        if not self.result:
            return

        columns = list(map(lambda x: result_fetch(self.result, x), range(self.result.ncols)))
        name = (make_string(rcol.name) for rcol in columns)
        type_code = (monet_c_type_map[rcol.type].sql_type for rcol in columns)
        display_size = repeat(None)
        internal_size = repeat(None)
        precision = repeat(None)
        scale = repeat(None)
        null_ok = repeat(None)
        descriptions = list(zip(name, type_code, display_size, internal_size, precision, scale, null_ok))
        return [Description(*i) for i in descriptions]

    def execute(
            self, query: str,
            args: parameters_type = None,
            cursor: Optional[Type['Cursor']] = None,
            paramstyle: str = "qmark"
    ) -> 'Cursor':
        """
        Execute a SQL query

        This is a nonstandard and SQLite compatible shortcut that creates a cursor object by calling the cursor()
        method, calls the cursor’s execute() method with the parameters given, and returns the cursor.

        Args:
            query: The SQL query to execute
            args:  The optional SQL query arguments
            cursor: An optional Cursor object
            paramstyle: The style of the args, can be qmark, numeric, format or pyformat

        Returns:
            A new cursor.
        """
        cur = self.cursor(factory=cursor).execute(query, args, paramstyle)
        self.consistent = True
        return cur

    def executemany(
            self,
            query: str,
            args_seq: Union[Iterator, Iterable[parameters_type]],
            cursor: Optional[Type['Cursor']] = None
    ) -> 'Cursor':
        """
        Prepare a database query and then execute it against all parameter sequences or mappings found in the
        sequence seq_of_parameters.

        This is a nonstandard and SQLite compatible shortcut that creates a cursor object by calling the cursor()
        method, calls the cursor’s execute() method with the parameters given, and returns the cursor.

        Args:
            query: The SQL query to execute
            args_seq:  The optional SQL query arguments
            cursor: A Cursor class

        Returns:
            A new cursor instance of the supplied cursor class
        """
        cur = self.cursor(factory=cursor)
        for args in args_seq:
            cur.execute(query, args)
        return cur

    def commit(self, *args, **kwargs) -> 'Cursor':
        self._check()
        return self.execute("COMMIT")

    def close(self, *args, **kwargs) -> None:
        if not hasattr(self, '_internal'):
            return

        if self._internal:
            self._internal.close()
        self._internal = None

    def cursor(self, factory: Optional[Type['Cursor']] = None) -> 'Cursor':
        """
        Create a new cursor.

        Args:
            factory: An optional factory. If supplied, this must be a callable returning an instance of Cursor or its
                     subclasses.

        Returns:
            a new cursor.
        """
        self._check()

        if not factory:
            from monetdbe.cursors import Cursor  # type: ignore[attr-defined]
            factory = Cursor

        cursor = factory(con=self)  # type: ignore[misc]
        if not cursor:
            raise TypeError

        return cursor

    def executescript(self, sql_script: str):
        self._check()
        for query in sql_script.split(';'):
            query = query.strip()
            if query:
                self.execute(query)

    def set_authorizer(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def backup(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def iterdump(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def create_collation(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def create_aggregate(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def set_progress_handler(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def set_trace_callback(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def create_function(self, *args, **kwargs):
        self._check()
        raise NotImplemented

    def rollback(self, *args, **kwargs):
        """
        Rolls back the current transaction.
        """
        self._check()
        self.execute("ROLLBACK")
        self.consistent = False

    @property
    def in_transaction(self):
        self._check()
        return self._internal.in_transaction()

    def set_autocommit(self, value: bool) -> None:
        """
        Set the connection to auto-commit mode.

        Args:
            value: a boolean value
        """
        self._check()
        return self._internal.set_autocommit(value)  # type: ignore

    def read_csv(self, table, *args, **kwargs):
        return self.cursor().read_csv(table, *args, **kwargs)

    def write_csv(self, table, *args, **kwargs):
        return self.cursor().write_csv(table, *args, **kwargs)

    def cleanup_result(self):
        if self.result and self._internal:
            self._internal.cleanup_result(self.result)
            self.result = None

    def query(self, query: str, make_result: bool = False) -> Tuple[Optional[Any], int]:
        """
        Execute a query directly on the connection.

        You probably don't want to use this. usually you use a cursor to execute queries.
        """
        self._check()
        return self._internal.query(query, make_result)  # type: ignore[union-attr]

    def prepare(self, operation: str):
        self._check()
        return self._internal.prepare(operation)  # type: ignore[union-attr]

    def cleanup_statement(self, statement: str) -> None:
        self._check()
        self._internal.cleanup_statement(statement)  # type: ignore[union-attr]

    def append(self, table: str, data: Mapping[str, np.ndarray], schema: str = 'sys') -> None:
        self._check()
        self._internal.append(table, data, schema)  # type: ignore[union-attr]

    # these are required by the python DBAPI
    Warning = exceptions.Warning
    Error = exceptions.Error
    InterfaceError = exceptions.InterfaceError
    DatabaseError = exceptions.DatabaseError
    DataError = exceptions.DataError
    OperationalError = exceptions.OperationalError
    IntegrityError = exceptions.IntegrityError
    InternalError = exceptions.InternalError
    ProgrammingError = exceptions.ProgrammingError
    NotSupportedError = exceptions.NotSupportedError



import logging
from re import compile, DOTALL

from monetdbe._lowlevel import ffi
from monetdbe._cffi.types_ import char_p

from monetdbe import exceptions

_logger = logging.getLogger(__name__)

error_matcher = compile(pattern=r"^(?P<exception_type>.*):(?P<namespace>.*):(?P<code>.*)!(?P<msg>.*)$", flags=DOTALL)

other_matchers = (
    ('07001', compile(pattern=r"^MALException:monetdbe.monetdbe_bind:(?P<msg>.*)", flags=DOTALL)),
    ('07001', compile(pattern=r"^MALException:monetdbe.monetdbe_execute:(?P<msg>.*)", flags=DOTALL)),
)

# MonetDB error codes
errors = {
    '2D000': exceptions.IntegrityError,  # COMMIT: failed
    '40000': exceptions.IntegrityError,  # DROP TABLE: FOREIGN KEY constraint violated
    '40002': exceptions.IntegrityError,  # INSERT INTO: UNIQUE constraint violated
    '42000': exceptions.OperationalError,  # SELECT: identifier 'asdf' unknown
    '42S02': exceptions.OperationalError,  # no such table
    '45000': exceptions.OperationalError,  # 'Result set construction failed'
    'M0M29': exceptions.IntegrityError,  # The code monetdb emitted before Jun2020
    '25001': exceptions.OperationalError,  # START TRANSACTION: cannot start a transaction within a transaction
    '07001': exceptions.ProgrammingError,  # Parameter .* not bound to a value
}


def check_error(raw: char_p) -> None:
    """
    Raises:
         exceptions.Error: or subclass in case of error, which exception depends on the error type.
    """
    if not raw:
        return

    decoded = ffi.string(raw).decode()
    _logger.error(decoded)
    match = error_matcher.match(decoded)

    if not match:
        for error_code, other_matcher in other_matchers:
            other_match = other_matcher.match(decoded)
            if other_match:
                exception = errors[error_code]
                msg = other_match.groups()[0]
                break
        else:
            # the error string is in an unknown format
            exception = exceptions.DatabaseError
            msg = decoded
    else:
        _, _, code, msg = match.groups()
        exception = errors.get(code, exceptions.DatabaseError)

    raise exception(msg)



class TestSimpleDBAPI:
    def test_regular_selection(self, monetdbe_cursor):
        monetdbe_cursor.execute('SELECT * FROM integers')
        result = monetdbe_cursor.fetchall()
        assert result == [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (None,)], "Incorrect result returned"



from decimal import Decimal
from typing import List, Optional, Callable, Union, Any, Mapping
from typing import NamedTuple

import numpy as np

from monetdbe._lowlevel import lib, ffi
from monetdbe._cffi.types_ import monetdbe_column, char_p
from monetdbe.converters import converters
from monetdbe.exceptions import ProgrammingError
from monetdbe.pythonize import py_date, py_time, py_timestamp


def make_string(blob: char_p) -> str:
    if blob:
        return ffi.string(blob).decode()
    else:
        return ""


def make_blob(blob: char_p) -> str:
    if blob:
        return ffi.string(blob.data[0:blob.size])
    else:
        return ""


def py_float(data: char_p) -> float:
    """
    Convert a monetdb FFI float to a python float
    """
    if 'FLOAT' in converters:  # type: ignore
        return converters['FLOAT'](data)
    elif 'DOUBLE' in converters:  # type: ignore
        return converters['DOUBLE'](data)
    else:
        return data


class MonetdbTypeInfo(NamedTuple):
    c_type: int
    sql_type: Optional[str]
    numpy_type: np.dtype
    c_string_type: str
    py_converter: Optional[Callable]
    null_value: Optional[Union[int, float]]


#  monetdb C type, SQL type, numpy type, Cstringtype, pyconverter, null value, comment
type_infos: List[MonetdbTypeInfo] = [
    MonetdbTypeInfo(lib.monetdbe_bool, "boolean", np.dtype(np.bool_), "bool", None, None),
    MonetdbTypeInfo(lib.monetdbe_int8_t, "tinyint", np.dtype(np.int8), "int8_t", None, np.iinfo(np.int8).min),  # type: ignore
    MonetdbTypeInfo(lib.monetdbe_int16_t, "smallint", np.dtype(np.int16), "int16_t", None, np.iinfo(np.int16).min),  # type: ignore
    MonetdbTypeInfo(lib.monetdbe_int32_t, "int", np.dtype(np.int32), "int32_t", None, np.iinfo(np.int32).min),  # type: ignore
    MonetdbTypeInfo(lib.monetdbe_int64_t, "bigint", np.dtype(np.int64), "int64_t", None, np.iinfo(np.int64).min),  # type: ignore
    MonetdbTypeInfo(lib.monetdbe_size_t, None, np.dtype(np.uint64), "size_t", None, None),  # used by monetdb internally
    MonetdbTypeInfo(lib.monetdbe_float, "real", np.dtype(np.float32), "float", py_float, np.finfo(np.float32).min),
    MonetdbTypeInfo(lib.monetdbe_double, "float", np.dtype(np.float64), "double", py_float, np.finfo(np.float64).min),
    MonetdbTypeInfo(lib.monetdbe_str, "string", np.dtype('=O'), "str", make_string, None),
    MonetdbTypeInfo(lib.monetdbe_blob, "blob", np.dtype('=O'), "blob", make_blob, None),
    MonetdbTypeInfo(lib.monetdbe_date, "date", np.dtype('=O'), "date", py_date, None),
    MonetdbTypeInfo(lib.monetdbe_time, "time", np.dtype('=O'), "time", py_time, None),
    MonetdbTypeInfo(lib.monetdbe_timestamp, "timestamp", np.dtype('=O'), "timestamp", py_timestamp, None),
]

numpy_type_map: Mapping[np.dtype, MonetdbTypeInfo] = {i.numpy_type: i for i in type_infos}
monet_c_type_map: Mapping[int, MonetdbTypeInfo] = {i.c_type: i for i in type_infos}


def numpy_monetdb_map(numpy_type: np.dtype):
    if numpy_type.kind in ('i', 'f'):  # type: ignore
        return numpy_type_map[numpy_type]
    raise ProgrammingError("append() only support int and float family types")


from monetdbe._cffi.branch import newer_then_jul2021
if newer_then_jul2021:
    def extract(rcol: monetdbe_column, r: int, text_factory: Optional[Callable[[str], Any]] = None):
        """
        Extracts values from a monetdbe_column.

        The text_factory is optional, and wraps the value with a custom user supplied text function.
        """

        type_info = monet_c_type_map[rcol.type]
        col = ffi.cast(f"monetdbe_column_{type_info.c_string_type} *", rcol)
        if col.is_null(col.data + r):
            return None
        else:
            col_data = col.data[r]
            if rcol.sql_type.name != ffi.NULL and ffi.string(rcol.sql_type.name).decode() == 'decimal':
                col_data = Decimal(col_data) / (Decimal(10) ** rcol.sql_type.scale)
            if type_info.py_converter:
                result = type_info.py_converter(col_data)
                if rcol.type == lib.monetdbe_str and text_factory:
                    return text_factory(result)
                return result
            return col_data
else:
    def extract(rcol: monetdbe_column, r: int, text_factory: Optional[Callable[[str], Any]] = None):
        """
        Extracts values from a monetdbe_column.
        The text_factory is optional, and wraps the value with a custom user supplied text function.
        """
        type_info = monet_c_type_map[rcol.type]
        col = ffi.cast(f"monetdbe_column_{type_info.c_string_type} *", rcol)
        if col.is_null(col.data + r):
            return None
        else:
            if type_info.py_converter:
                result = type_info.py_converter(col.data[r])
                if rcol.type == lib.monetdbe_str and text_factory:
                    return text_factory(result)
                return result
            return col.data[r]



import pytest

import monetdbe


class TestShutdown:
    def test_commited_on_restart(self, monetdbe_cursor_autocommit):
        context = monetdbe_cursor_autocommit
        cursor = context.cursor
        cursor.transaction()
        cursor.execute('CREATE TABLE integers (i INTEGER)')
        cursor.executemany('INSERT INTO integers VALUES (%s)', [[x] for x in range(3)])
        cursor.execute('SELECT * FROM integers')
        result = cursor.fetchall()
        assert result == [(0,), (1,), (2,)], "Incorrect result returned"
        cursor.commit()
        context.connection.close()

        context.connection = monetdbe.make_connection(context.dbfarm)
        cursor = context.connection.cursor()
        cursor.execute('SELECT * FROM integers')
        assert result == [(0,), (1,), (2,)], "Incorrect result returned"

    def test_transaction_aborted_on_shutdown(self, monetdbe_cursor_autocommit):
        context = monetdbe_cursor_autocommit
        cursor = context.cursor
        cursor.transaction()
        cursor.execute('CREATE TABLE integers (i INTEGER)')
        cursor.executemany('INSERT INTO integers VALUES (%s)', [[x] for x in range(3)])
        cursor.execute('SELECT * FROM integers')
        result = cursor.fetchall()
        assert result == [(0,), (1,), (2,)], "Incorrect result returned"
        context.connection.close()

        context.connection = monetdbe.make_connection(context.dbfarm)
        cursor = context.connection.cursor()
        with pytest.raises(monetdbe.DatabaseError):
            cursor.execute('SELECT * FROM integers')

    def test_many_shutdowns(self, monetdbe_cursor_autocommit):
        context = monetdbe_cursor_autocommit
        cursor = context.cursor
        for i in range(10):
            cursor.transaction()
            cursor.execute('CREATE TABLE integers (i INTEGER)')
            cursor.executemany('INSERT INTO integers VALUES (%s)', [[x] for x in range(10)])
            cursor.execute('SELECT MIN(i * 3 + 5) FROM integers')
            result = cursor.fetchall()
            assert result == [(5,)], "Incorrect result returned"
            context.connection.close()

            context.connection = monetdbe.make_connection(context.dbfarm)
            connection = context.connection
            connection.set_autocommit(True)
            cursor = connection.cursor()

    @pytest.mark.skip("this is in conflict with the DB API and SQLite test suite")
    def test_fetchone_without_executing_raises(self, monetdbe_empty_cursor):
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe_empty_cursor.fetchone()

    @pytest.mark.skip("this is in conflict with the DB API and SQLite test suite")
    def test_fetchall_without_executing_raises(self, monetdbe_empty_cursor):
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe_empty_cursor.fetchall()

    def test_fetchnumpy_without_executing_raises(self, monetdbe_empty_cursor):
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe_empty_cursor.fetchnumpy()

    def test_fetchdf_without_executing_raises(self, monetdbe_empty_cursor):
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe_empty_cursor.fetchdf()

    def test_execute_with_closed_cursor_raises(self, monetdbe_empty_cursor):
        monetdbe_empty_cursor.close()
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe_empty_cursor.execute("SELECT * FROM _tables")

    def test_fetchmany(self, monetdbe_cursor):
        monetdbe_cursor.execute("SELECT * FROM integers")

        counter = 0
        while counter < 10:
            r = monetdbe_cursor.fetchmany(2)
            assert len(r) == 2
            counter += len(r)

        assert counter == 10

    def test_fetchmany_without_explicit_size(self, monetdbe_cursor):
        assert monetdbe_cursor.arraysize == 1, "Incorrect default value for cursor.arraysize"
        monetdbe_cursor.arraysize = 2
        monetdbe_cursor.execute("SELECT * FROM integers")

        counter = 0
        while counter < 10:
            r = monetdbe_cursor.fetchmany()
            assert len(r) == 2
            counter += len(r)

        assert counter == 10

    @pytest.mark.skip("we dont support scrolling")
    def test_scroll(self, monetdbe_cursor):
        monetdbe_cursor.execute("SELECT * FROM integers")
        monetdbe_cursor.scroll(5)

        x = monetdbe_cursor.fetchone()
        assert x[0] == 6

    @pytest.mark.skip("we dont support scrolling")
    def test_scroll_raises_for_incorrect_mode(self, monetdbe_cursor):
        monetdbe_cursor.execute("SELECT * FROM integers")
        with pytest.raises(monetdbe.ProgrammingError):
            monetdbe_cursor.scroll(5, mode='abc')

    @pytest.mark.skip("we dont support scrolling")
    def test_scroll_raises_for_out_of_bounds_offset(self, monetdbe_cursor):
        monetdbe_cursor.execute("SELECT * FROM integers")
        with pytest.raises(IndexError):
            monetdbe_cursor.scroll(20)

    @pytest.mark.xfail(reason="We do not implement correctly the iterator protocol for py27")
    def test_cursor_iteration_protocol(self, monetdbe_cursor):
        monetdbe_cursor.execute("SELECT * FROM integers WHERE i IS NOT NULL")

        counter = 0
        for i in monetdbe_cursor:
            assert i[0] == counter
            counter += 1

        assert counter == 10
    # TODO: rewrite this one
    # def test_use_old_cursor(self, monetdbe_cursor):
    #     self.connection.close()

    #     self.connection = monetdbe.connect(self.dbfarm)
    #     if not PY26:
    #         with self.assertRaises(monetdbe.ProgrammingError):
    #             monetdbe_cursor.execute('SELECT * FROM integers')



from setuptools import find_packages, setup
from sys import platform

with open("README.md", "r") as fh:
    long_description = fh.read()

tests_require = [
    'pytest',
    'mypy',
    'pycodestyle',
    'data-science-types',
    'types-setuptools',
    'types-pkg_resources',
    'types-Jinja2',
    'typing-extensions',
]

extras_require = {
    'test': tests_require,
    'doc': ['sphinx', 'sphinx_rtd_theme'],

}

packages = find_packages(exclude=['tests', 'tests.test_lite'])

if platform == 'win32':
    package_data = {"monetdbe": ["*.dll"]}
else:
    package_data = {}


setup(
    name="monetdbe",
    version="0.10.1",
    author="Gijs Molenaar",
    author_email="gijs@pythonic.nl",
    description="MonetDBe - the Python embedded MonetDB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/monetdBSolutions/MonetDBe-Python/",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Topic :: Database",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: Database :: Front-Ends",
    ],
    python_requires='>=3.6',
    setup_requires=["cffi>=1.0.0", "Jinja2"],
    extras_require=extras_require,
    cffi_modules=["monetdbe/_cffi/builder.py:ffibuilder"],
    install_requires=["cffi>=1.0.0", "numpy", "pandas"],
    tests_require=tests_require,
    test_suite="tests",
    package_data=package_data,
)



import pkg_resources

try:
    __version__ = pkg_resources.require("monetdbe")[0].version  # type: str
except pkg_resources.DistributionNotFound:
    __version__ = "0.0"

version = __version__
monetdbe_version = __version__
version_info = tuple([int(x) for x in __version__.split(".")])
monetdbe_version_info = version_info



from monetdbe.compat import (
    create,
    make_connection,
    init,
    insert,
    shutdown,
    sql
)

from monetdbe.converters import register_adapter
from monetdbe.monetize import PrepareProtocol
from monetdbe.row import Row
from monetdbe.version import monetdbe_version_info
from monetdbe.cursors import Cursor  # type: ignore[attr-defined]
from monetdbe.connection import Connection

from monetdbe.dbapi2 import (
    connect,
    Timestamp,
    OptimizedUnicode,
    apilevel,
    paramstyle,
    threadsafety,
    Binary,
    TimestampFromTicks,
    DateFromTicks,
    TimeFromTicks,
    Date,
    Time,

)

from monetdbe.exceptions import (
    OperationalError,
    Error,
    DataError,
    DatabaseError,
    IntegrityError,
    InternalError,
    InterfaceError,
    StandardError,
    Warning,
    NotSupportedError,
    ProgrammingError
)



import unittest
from sys import platform
import numpy as np
from monetdbe._lowlevel import lib
from monetdbe import connect
from monetdbe.exceptions import ProgrammingError


class TestCffi(unittest.TestCase):
    def test_cffi(self):
        if platform == 'win32':
            # windows has 128 bit disabled
            self.assertEqual(lib.monetdbe_type_unknown, 13)
        else:
            self.assertEqual(lib.monetdbe_type_unknown, 14)

    def test_append_too_many_columns(self):
        with connect() as con:
            con.execute("CREATE TABLE test (i int)")
            data = {'i': np.array([1, 2, 3]), 'j': np.array([1, 2, 3])}
            with self.assertRaises(ProgrammingError):
                con._internal.append(table='test', data=data)

    def test_append_too_little_columns(self):
        with connect() as con:
            con.execute("CREATE TABLE test (i int, j int)")
            data = {'i': np.array([1, 2, 3])}
            with self.assertRaises(ProgrammingError):
                con._internal.append(table='test', data=data)

    def test_append_wrong_type(self):
        with connect() as con:
            con.execute("CREATE TABLE test (i int)")
            data = {'i': np.array([0.1, 0.2, 0.3], dtype=np.float32)}
            with self.assertRaises(ProgrammingError):
                con._internal.append(table='test', data=data)

    def test_append_wrong_size(self):
        with connect() as con:
            con.execute("CREATE TABLE test (i int)")  # SQL int is 32 bit
            data = {'i': np.array([1, 2, 3], dtype=np.int64)}
            with self.assertRaises(ProgrammingError):
                con._internal.append(table='test', data=data)

    def test_append_supported_types(self):
        with connect() as con:
            con.execute("CREATE TABLE test (t tinyint, s smallint, i int, b bigint, r real, f float)")
            con.execute(
                """
                INSERT INTO test VALUES (2^8,  2^16,  2^32,  2^64,  0.12345,  0.123456789),
                                        (NULL, NULL,  NULL,  NULL,  NULL,     NULL),
                                        (0,    0,     0,     0,     0.0,      0.0),
                                        (-2^8, -2^16, -2^32, -2^64, -0.12345, -0.123456789)
                """
            )
            data = con.execute("select * from test").fetchnumpy()
            con._internal.append(schema='sys', table='test', data=data)
            con.cursor().insert(table='test', values=data)

    def test_append_unsupported_types(self):
        with connect() as con:
            con.execute("CREATE TABLE test (s string, b blob, d date, t time, ts timestamp)")
            con.execute(
                """
                INSERT INTO test VALUES ('hi',    '01020308', '2020-01-02', '10:20:30', '2020-01-02 10:20:30' ),
                                        ('World', NULL,       NULL,         NULL,       NULL )
                """
            )

            data = con.execute("select * from test").fetchnumpy()
            with self.assertRaises(con.ProgrammingError):
                con._internal.append(schema='sys', table='test', data=data)
            con.cursor().insert(table='test', values=data)

    def test_append_blend(self):
        with connect() as con:
            con.execute("CREATE TABLE test (i int, f float, s string, ts timestamp)")
            con.execute(
                """
                INSERT INTO test VALUES (1, 1.2, 'bla', '2020-01-02 10:20:30'),
                                        (NULL, NULL, NULL, NULL)
                """
            )

            data = con.execute("select * from test").fetchnumpy()
            with self.assertRaises(con.ProgrammingError):
                con._internal.append(schema='sys', table='test', data=data)
            con.cursor().insert(table='test', values=data)

    def test_get_columns(self):
        with connect() as con:
            con.execute("CREATE TABLE test (i int)")
            con.execute("INSERT INTO test VALUES (1)")
            result = list(con._internal.get_columns(table='test'))
            self.assertEqual(result, [('i', 3)])



from typing import Any, Union
from decimal import Decimal
import datetime
from monetdbe._lowlevel import ffi


def monetdbe_decimal_to_bte(data: int) -> ffi.CData:
    return ffi.new("int8_t *", data)


def monetdbe_decimal_to_sht(data: int) -> ffi.CData:
    return ffi.new("int16_t *", data)


def monetdbe_decimal_to_int(data: int) -> ffi.CData:
    return ffi.new("int *", data)


def monetdbe_decimal_to_lng(data: int) -> ffi.CData:
    return ffi.new("int64_t *", data)


def monetdbe_int(data: int) -> ffi.CData:
    if data > 2 ** 32:
        return ffi.new("int64_t *", data)
    else:
        return ffi.new("int *", data)


def bind_str(data: str) -> ffi.CData:
    return ffi.new("char[]", str(data).encode())


def bind_float(data: float) -> ffi.CData:
    return ffi.new("double *", data)


def bind_bytes(data: bytes) -> bytes:
    return f"{data.hex()}".encode()


def bind_memoryview(data: memoryview) -> ffi.CData:
    struct = ffi.new("monetdbe_data_blob *")
    bytes_ = data.tobytes()
    struct.size = len(bytes_)
    struct.data = ffi.new("char []", bytes_)
    return struct


def bind_datetime(data: datetime.datetime) -> ffi.CData:
    struct = ffi.new("monetdbe_data_timestamp *")
    struct.date.day = data.day
    struct.date.month = data.month
    struct.date.year = data.year
    struct.time.ms = int(data.microsecond / 1000)
    struct.time.seconds = data.second
    struct.time.minutes = data.minute
    struct.time.hours = data.hour
    return struct


def bind_time(data: datetime.time) -> ffi.CData:
    struct = ffi.new("monetdbe_data_time *")
    struct.ms = int(data.microsecond / 1000)
    struct.seconds = data.second
    struct.minutes = data.minute
    struct.hours = data.hour
    return struct


def bind_date(data: datetime.date) -> ffi.CData:
    struct = ffi.new("monetdbe_data_date *")
    struct.day = data.day
    struct.month = data.month
    struct.year = data.year
    return struct


def bind_timedelta(data: datetime.timedelta) -> ffi.CData:
    struct = ffi.new("monetdbe_data_time *")
    struct.ms = int(data.microseconds / 1000)
    struct.seconds = data.seconds
    struct.minutes = 0
    struct.hours = 0
    return struct


map_ = {
    int: monetdbe_int,
    datetime.datetime: bind_datetime,
    datetime.time: bind_time,
    datetime.date: bind_date,
    datetime.timedelta: bind_timedelta,
    str: bind_str,
    float: bind_float,
    bytes: bind_bytes,
    memoryview: bind_memoryview,
}


def prepare_bind(data: Any) -> Union[str, ffi.CData]:
    try:
        return map_[type(data)](data)  # type: ignore[operator]
    except KeyError:
        for type_, func in map_.items():
            if isinstance(data, type_):
                return func(data)  # type: ignore[operator]
        raise NotImplementedError(f"bind converting for type {type(data)}")
